# -*- coding: utf-8 -*-

"""
Module for mapping Twitter to common LBSN Structure.
"""

# pylint: disable=no-member

import logging
import re

import shapely.geometry as geometry
from lbsnstructure import lbsnstructure_pb2 as lbsn
from shapely.geometry.polygon import Polygon

from lbsntransform.tools.helper_functions import HelperFunctions as HF

MAPPING_ID = 3
class importer():
    """ Provides mapping function from Twitter endpoints to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "Twitter"
    ORIGIN_ID = 3

    def __init__(self,
                 disableReactionPostReferencing=False,
                 geocodes=False,
                 mapFullRelations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=None,
                 min_geoaccuracy=None):
        # We're dealing with Twitter in this class,
        # lets create the OriginID globally
        # this OriginID is required for all CompositeKeys
        origin = lbsn.Origin()
        origin.origin_id = lbsn.Origin.TWITTER
        self.origin = origin
        # this is where all the data will be stored
        # self.lbsn_records = LBSNRecordDicts()
        self.lbsn_records = []
        self.null_island = 0
        self.log = logging.getLogger('__main__')  # logging.getLogger()
        self.disable_reaction_post_referencing = disableReactionPostReferencing
        self.map_full_relations = mapFullRelations
        self.geocodes = geocodes
        self.map_reactions = map_reactions
        self.ignore_non_geotagged = ignore_non_geotagged
        self.ignore_sources_set = ignore_sources_set
        self.min_geoaccuracy = min_geoaccuracy
        self.skipped_low_geoaccuracy = 0
        self.skipped_ignore_list = 0

    def get_skipped_geoaccuracy(self):
        """Get count of records skipped due to low geoaccuracy"""
        return self.skipped_low_geoaccuracy

    def get_skipped_ignorelist(self):
        """Get count of records skipped due to ignore list"""
        return self.skipped_ignore_list

    def parse_json_record(self, json_string_dict, input_lbsn_type=None):
        """Will parse Twitter json retrieved from Twitter API,
        returns a list of LBSN records.


        Fully extracting Twitter json's to flat relational db structure
        is challenging because Twitter json's may consist of deeply nested
        structures, which can include many LBSN record entities, e.g.:
        - the lbsn.Post itself
        - the lbsn.User who posted, and its attributes
        - Coordinates, Places, Cities, Countries linked to the post
        - lbsn.Language of the post
        - shared or retweeted Posts and their attribues (
            Users, Places, Cities etc.)
        - mentioned users in the post ("@-mentions")
        - special jsons retrieved from other API endpoints, e.g.
            groups of users etc.

        This methods tries to do all of this automatically, but default
        values may need adjustment for specific cases. All extracted
        LBSN records are added subsequently to self.lbsn_records and
        returned finally as a single list of records in this method. This
        guarantees that db-key-relations are acknowledged when submitting
        records to db. The order of LBSN record type extraction
        follows the order of db inserts
        """
        # clear any records from previous run
        self.lbsn_records.clear()
        # decide if main object is post or user json
        if input_lbsn_type and input_lbsn_type in ('friendslist',
                                                   'followerslist'):
            for user, related_user_list in json_string_dict.items():
                user_record = HF.new_lbsn_record_with_id(
                    lbsn.User(), str(user), self.origin)
                self.lbsn_records.append(user_record)
                self.extract_related_users(related_user_list,
                                           input_lbsn_type, user_record)
        elif (input_lbsn_type and input_lbsn_type == 'profile') \
                or 'screen_name' in json_string_dict:
            # user
            user_record = self.extract_user(json_string_dict)
            self.lbsn_records.append(user_record)
            # sys.exit(f'lbsn.Post record: {text_format.MessageToString(userRecord,
            #                                                      as_utf8=True)}')
            if not user_record.is_private:
                # if user profile is private, we cannot access posts
                user_status = None
                if 'status' in json_string_dict:
                    user_status = json_string_dict.get('status')
                elif 'quoted_status' in json_string_dict:
                    user_status = json_string_dict.get('quoted_status')
                elif 'retweeted_status' in json_string_dict:
                    user_status = json_string_dict.get('retweeted_status')
                # in case user status is available
                if user_status:
                    self.parse_json_post(
                        user_status, user_pkey=user_record.pkey)
        else:
            # otherwise, parse post
            self.parse_json_post(json_string_dict)

        # finally, return list of all extracted records
        return self.lbsn_records

    def extract_related_users(
            self, related_user_list, input_lbsn_type, user_record):
        """Extract related users from user list"""
        for related_user in related_user_list:
            related_record = HF.new_lbsn_record_with_id(lbsn.User(),
                                                        str(related_user),
                                                        self.origin)
            self.lbsn_records.append(related_record)
            # note the switch of order here,
            # direction is important for 'isConnected',
            # and the different list each give us a
            # different view on this relationship
            if input_lbsn_type == 'friendslist':
                relationship_record =\
                    HF.new_lbsn_relation_with_id(lbsn.Relationship(),
                                                 user_record.pkey.id,
                                                 related_record.pkey.id,
                                                 self.origin)
            elif input_lbsn_type == 'followerslist':
                relationship_record = \
                    HF.new_lbsn_relation_with_id(lbsn.Relationship(),
                                                 related_record.pkey.id,
                                                 user_record.pkey.id,
                                                 self.origin)
            relationship_record.relationship_type = \
                lbsn.Relationship.isCONNECTED
            self.lbsn_records.add_relationship_to_dict(
                relationship_record)

    def parse_json_post(self, json_string_dict, user_pkey=None):
        """Extract json post retrieved from Twitter API

        The process is nested, but pretty linear:
        1. Extract all relevant lbsn.Post Attributes
           1.a extract post coordinates
           1.b extract user attributes
           1.c extract place attributes
        (poi, city, neigborhood, admin, country)
           1.d extract extract extended tweet,
        if available, and extended entities, if available
        2. decide if post is reaction
        (reply, quote, share, see https://developer.twitter.com/
        en/docs/tweets/data-dictionary/overview/entities-object.html)
        3. if post is reaction, copy reduced reaction
        attributes from extracted lbsn.Post
        4. add post/reaction to recordDict
        5. process all referenced posts
           5.a Retweet(=Share) and Quote Tweets are special kinds
        of Tweets that contain the original Tweet as an embedded object.
           5.b Retweets have a top-level "retweeted_status"
        object, and Quoted Tweets have a "quoted_status" object
        process tweet-post object

        Note: one input record may contain many lbsn records
        therefore, records are first added to self.lbsn_records
        to be later returned together
        """
        post_record = self.extract_post(
            json_string_dict, user_pkey)

        if not post_record:
            # in case no post record has been extracted
            # (e.g. non_geotagged clause)
            return
        # Assignment Step
        # check if post is reaction to other post
        # reaction means: reduced structure compared to post;
        # reactions often include the complete original post,
        # therefore nested processing necessary
        if HF.is_post_reaction(json_string_dict):
            if self.map_reactions is False:
                return
            post_reaction_record = self.map_postrecord_to_postreactionrecord(
                post_record)
            refuser_pkey = None
            if 'quoted_status' in json_string_dict:
                # Note: Quote is both: Share & Reply
                if 'user' not in json_string_dict.get('quoted_status'):
                    refuser_pkey = \
                        HF.substitute_referenced_user(json_string_dict,
                                                      self.origin,
                                                      self.log)
                post_reaction_record.reaction_type = lbsn.PostReaction.QUOTE
                ref_post_record = self.extract_post(
                    json_string_dict.get('quoted_status'))
            elif 'retweeted_status' in json_string_dict:
                # Note: No retweets are available when data is queried
                # using Bounding Box because of Geo-Tweet limitation:
                # "Note that native Retweets are not matched by this
                # parameter. While the original Tweet may have a location,
                # the Retweet will not"
                # see https://developer.twitter.com/en/docs/
                # tweets/filter-realtime/guides/basic-stream-parameters.html
                if 'user' not in json_string_dict.get('retweeted_status'):
                    # Current issue with Twitter search: the retweeting
                    # user is not returned in retweeted_status
                    # but we can get this from other information,
                    # such as user_mentions field from the retweet
                    # https://twittercommunity.com/t/status-retweeted-
                    # status-quoted-status-user-missing-from-search-tweets-json-response/63355
                    refuser_pkey = \
                        HF.substitute_referenced_user(json_string_dict,
                                                      self.origin,
                                                      self.log)
                post_reaction_record.reaction_type = lbsn.PostReaction.SHARE
                retweet_post = json_string_dict.get('retweeted_status')
                ref_post_record = self.extract_post(retweet_post, refuser_pkey)

            elif json_string_dict.get('in_reply_to_status_id_str'):
                # if reply, original tweet is not available (?)
                post_reaction_record.reaction_type = lbsn.PostReaction.COMMENT
                ref_post_record = \
                    HF.new_lbsn_record_with_id(
                        lbsn.Post(), json_string_dict.get(
                            'in_reply_to_status_id_str'),
                        self.origin)
                ref_user_record = \
                    HF.new_lbsn_record_with_id(
                        lbsn.User(),
                        json_string_dict.get(
                            'in_reply_to_user_id_str'),
                        self.origin)
                ref_user_record.user_name = json_string_dict.get(
                    'in_reply_to_screen_name')  # Needs to be saved
                self.lbsn_records.append(ref_user_record)
                ref_post_record.user_pkey.CopyFrom(ref_user_record.pkey)

            # add referenced post pkey to reaction
            if not self.disable_reaction_post_referencing:
                post_reaction_record.referencedPost_pkey.CopyFrom(
                    ref_post_record.pkey)
                # ToDo: if a Reaction refers to another
                # reaction (Information Spread)
                # This information is currently not
                # [available from Twitter](https://developer.twitter.com/
                # en/docs/tweets/data-dictionary/overview/tweet-object):
                # "Note that retweets of retweets do not show
                # representations of the intermediary retweet [...]"
                # would be added to
                # postReactionRecord.referencedPostReaction_pkey
                if ref_post_record:
                    self.lbsn_records.append(ref_post_record)
            # add postReactionRecord to Dict
            self.lbsn_records.append(post_reaction_record)
        else:
            # otherwise add post to self.lbsn_records
            # which already includes all other entries (lbsn.User, lbsn.City, lbsn.Place etc.)
            self.lbsn_records.append(post_record)

    def extract_user(self, json_string_dict):
        """Extract lbsn.User from Twitter json"""
        user = json_string_dict
        user_record = HF.new_lbsn_record_with_id(
            lbsn.User(), user.get('id_str'), self.origin)
        # get additional information about the user, if available
        user_record.user_fullname = user.get('name')
        user_record.follows = user.get('friends_count')
        user_record.is_private = user.get('protected')
        user_record.followed = user.get('followers_count')
        user_bio = user.get('description')
        if user_bio:
            user_record.biography = user_bio
        user_record.user_name = user.get('screen_name')
        listed_count = user.get('listed_count')
        if listed_count:
            user_record.group_count = listed_count
        user_record.post_count = user.get('statuses_count')
        user_record.url = f'https://twitter.com/intent/user?user_id=' \
                          f'{user_record.pkey.id}'
        ref_user_language = lbsn.Language()
        ref_user_language.language_short = user.get('lang')
        user_record.user_language.CopyFrom(ref_user_language)
        user_location = user.get('location')
        if user_location:
            user_record.user_location = user_location
            if self.geocodes and user_record.user_location in self.geocodes:
                l_lat = self.geocodes[user_record.user_location][0]
                l_lng = self.geocodes[user_record.user_location][1]
                user_record.user_location_geom = "POINT(%s %s)" % (
                    l_lng, l_lat)
        # userGeoLocation = user.get('profile_location') # todo!
        user_record.liked_count = user.get('favourites_count')
        user_record.active_since.CopyFrom(
            HF.json_date_string_to_proto(user.get('created_at')))
        user_profile_image_url = user.get('profile_image_url')
        if not user_profile_image_url == f'http://abs.twimg.com/sticky/' \
                                         f'default_profile_images/' \
                                         f'default_profile_normal.png':
            user_record.profile_image_url = user_profile_image_url
        user_timezone = user.get('time_zone')
        if user_timezone:
            user_record.user_timezone = user_timezone
        user_utc_offset = user.get('utc_offset')
        if user_utc_offset:
            user_record.user_utc_offset = user_utc_offset
        # the following example demonstrates specific information
        # that cannot be extracted from twitter post data
        # deutscherBundestagGroup = \
        # HF.createNewLBSNRecord_with_id(lbsn.UserGroup(),
        #                               "MdB (Bundestag)",
        #                               self.origin)
        # userRecord.user_groups_member.append(
        #    deutscherBundestagGroup.pkey.id)
        # if self.mapFullRelations:
        #       relationshipRecord = \
        #       HF.createNewLBSNRelationship_with_id(lbsn.Relationship(),
        #                                            userRecord.pkey.id,
        #                                            deutscherBundestagGroup.pkey.id,
        #                                            self.origin)
        #       relationshipRecord.relationship_type = lbsn.Relationship.inGROUP
        #       self.lbsn_records.AddRelationshipToDict(relationshipRecord)
        # userRecord.user_groups_follows = []
        return user_record

    def extract_post(self, json_string_dict, user_pkey=None):
        """Returns tuple of lbsn.Post() and List of post_context_records

        e.g.:
            (lbsn.Post(), [lbsn.Country(), lbsn.City(), lbsn.Place(), lbsn.User()])
        """
        post_guid = json_string_dict.get('id_str')

        if not HF.check_notice_empty_post_guid(post_guid):
            return None, None
        post_record = HF.new_lbsn_record_with_id(lbsn.Post(),
                                                 post_guid,
                                                 self.origin)
        post_geoacc = None
        user_record = None
        user_info = json_string_dict.get('user')
        if user_info:
            # Get lbsn.Post/Reaction Details of lbsn.User
            user_record = self.extract_user(json_string_dict.get('user'))
        elif user_pkey:
            # userPkey is already available for posts that are statuses
            user_record = HF.new_lbsn_record_with_id(lbsn.User(),
                                                     user_pkey.id,
                                                     self.origin)
        if user_record:
            # self.lbsn_records.append(user_record)
            self.lbsn_records.append(user_record)
        else:
            self.log.warning(f'Record {self.lbsn_records.count_glob_total}: '
                             f'No lbsn.User record found for post: {post_guid} '
                             f'(post saved without userid)..')
            print(f'Record {self.lbsn_records.count_glob_total}', end='\r')

        # Some preprocessing for all types:
        post_coordinates = json_string_dict.get('coordinates')
        if post_coordinates:
            l_lng = post_coordinates.get('coordinates')[0]
            l_lat = post_coordinates.get('coordinates')[1]
            post_record.post_geoaccuracy = lbsn.Post.LATLNG
            post_record.post_latlng = "POINT(%s %s)" % (l_lng, l_lat)

        # Check if lbsn.Place is mentioned
        post_place_json = json_string_dict.get('place')
        if post_place_json:
            # we need some information from postRecord to create placeRecord
            # (e.g. user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord
            # attributes; therefore return both
            if user_record:
                user_lang = user_record.user_language
            else:
                user_lang = None
            place_record, \
                post_geoacc, \
                post_country = self.extract_place(post_place_json,
                                                  post_record.post_geoaccuracy,
                                                  user_lang)
            if not post_record.post_geoaccuracy:
                post_record.post_geoaccuracy = post_geoacc
            # postRecord.post_geoaccuracy = twitterPostAttributes.geoaccuracy
            # self.lbsn_records.append(place_record)
            self.lbsn_records.append(place_record)
            if post_country:
                post_record.country_pkey.CopyFrom(post_country.pkey)
            if isinstance(place_record, lbsn.City):
                post_record.city_pkey.CopyFrom(place_record.pkey)
            # either city or place, Twitter user cannot attach both (?)
            elif isinstance(place_record, lbsn.Place):
                post_record.place_pkey.CopyFrom(place_record.pkey)
            # substitute postRecord LatLng Coordinates from placeRecord,
            # if not already set
            if not post_record.post_latlng:
                # Note: this will also substitute lbsn.Country lat/lng in post
                # this information is also available by query of
                # country_guid in posts
                # use input arg min_geoaccuracy to exclude country geo-posts
                post_record.post_latlng = place_record.geom_center
        # if still no geoinformation, send post to Null-Island
        if not post_record.post_latlng:
            if self.ignore_non_geotagged is True:
                return None
            self.null_island += 1
            post_record.post_latlng = "POINT(%s %s)" % (0, 0)
        if self.min_geoaccuracy:
            if not HF.geoacc_within_threshold(post_record.post_geoaccuracy,
                                              self.min_geoaccuracy):
                self.skipped_low_geoaccuracy += 1
                return None
        # Process attributes of twitter post
        post_source = json_string_dict.get('source')
        if post_source:
            post_record.input_source = HF.cleanhtml(
                json_string_dict.get('source'))
            if self.ignore_sources_set and \
                    post_record.input_source in self.ignore_sources_set:
                # skip entry if in ignore list
                self.skipped_ignore_list += 1
                return None
        post_record.post_publish_date.CopyFrom(
            HF.json_date_string_to_proto(json_string_dict.get('created_at')))
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)
        post_record.post_quote_count = HF.value_count(
            json_string_dict.get('quote_count'))
        post_record.post_comment_count = HF.value_count(
            json_string_dict.get('reply_count'))
        post_record.post_share_count = HF.value_count(
            json_string_dict.get('retweet_count'))
        post_record.post_like_count = HF.value_count(
            json_string_dict.get('favorite_count'))
        post_record.post_url = f'https://twitter.com/statuses/{post_guid}'
        language_str = json_string_dict.get('lang')
        if language_str:
            post_language = lbsn.Language()
            post_language.language_short = json_string_dict.get('lang')
            post_record.post_language.CopyFrom(post_language)
        # If Extended_tweet object is available,
        # process entities and post_body (text) data from extended object
        is_truncated = json_string_dict.get('truncated')
        if is_truncated and 'extended_tweet' in json_string_dict:
            # if the "truncated" field is set to true,
            # and the "extended_tweet" object provides complete
            # "full_text" and "entities" Tweet metadata
            # Source for all data is extended object, if available
            json_string_dict = json_string_dict.get('extended_tweet')
            post_record.post_body = json_string_dict.get('full_text')
            # else:
            #    self.log.warning(f'Truncated but no extended_tweet:'
            #                     f'{json_string_dict}')
            #    input("Press Enter to continue... (entry will be skipped)")
            #    return None
        else:
            if 'full_text' in json_string_dict:
                post_record.post_body = json_string_dict.get('full_text')
            else:
                post_record.post_body = json_string_dict.get('text')
        # entities section always exists and includes meta information
        # such as hashtags or user_mentions
        entities_json = json_string_dict.get('entities')
        # extract hashtags
        hashtags_json = entities_json.get('hashtags')
        if hashtags_json:
            for hashtag in hashtags_json:  # iterate over the list
                post_record.hashtags.append(hashtag.get("text"))
        # Look for mentioned userRecords
        user_mentions_json = entities_json.get('user_mentions')
        if user_mentions_json:
            ref_user_records = HF.get_mentioned_users(user_mentions_json,
                                                      self.origin)
            # self.lbsn_records.append(ref_user_records)
            self.lbsn_records.append(ref_user_records)
            post_record.user_mentions_pkey.extend(
                [user_ref.pkey for user_ref in ref_user_records])
            if self.map_full_relations:
                self.extract_mentioned_users(
                    ref_user_records, user_record.pkey.id)
        # sometimes, extended_entities section exists and includes
        # additional information on media, but never hashtags or user_mentions
        # Since the media type metadata in the extended_entities section
        # correctly indicates the media type
        # (‘photo’, ‘video’ or ‘animated_gif’),
        # and supports up to 4 photos, it is the preferred metadata
        # source for native media. See:
        # https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/extended-entities-object.html#extended-entities-object
        if 'extended_entities' in json_string_dict:
            entities_json = json_string_dict.get('extended_entities')
        media_json = entities_json.get('media')
        if media_json:
            post_record.post_type = HF.assign_media_post_type(media_json)
        else:
            post_record.post_type = lbsn.Post.TEXT
        post_record.emoji.extend(HF.extract_emoji(post_record.post_body))
        # because standard print statement will produce escaped text,
        # we can use protobuf text_format to give us a human friendly
        # version of the text
        # log.debug(f'lbsn.Post record: '
        #           f'{text_format.MessageToString(postRecord, as_utf8=True)}')
        # log.debug(f'lbsn.Post record: {postRecord}')
        return post_record

    def extract_mentioned_users(self, ref_user_records, user_record_id):
        """Extract mentioned user from ref user records list"""
        for mentioned_user_record in ref_user_records:
            relation_record = \
                HF.new_lbsn_relation_with_id(lbsn.Relationship(),
                                             user_record_id,
                                             mentioned_user_record.pkey.id,
                                             self.origin)
            relation_record.relationship_type = \
                lbsn.Relationship.MENTIONS_USER
            self.lbsn_records.add_relationship_to_dict(
                relation_record)

    @classmethod
    def map_postrecord_to_postreactionrecord(cls, post_record):
        """Reduces lbsn.Post to lbsn.PostReaction record"""
        post_reaction_record = lbsn.PostReaction()
        post_reaction_record.pkey.CopyFrom(post_record.pkey)
        post_reaction_record.user_pkey.CopyFrom(post_record.user_pkey)
        post_reaction_record.reaction_latlng = post_record.post_latlng
        # better post_create_date, but not available from Twitter
        post_reaction_record.reaction_date.CopyFrom(
            post_record.post_publish_date)
        post_reaction_record.reaction_like_count = post_record.post_like_count
        post_reaction_record.reaction_content = post_record.post_body
        post_reaction_record.user_mentions_pkey.extend(
            post_record.user_mentions_pkey)
        return post_reaction_record

    def extract_place(
            self, postplace_json,
            post_geoaccuracy, user_language=None):
        """Extract lbsn.Place from twitter json"""
        place = postplace_json
        place_id = place.get('id')
        if not place_id:
            self.log.warning(f'No PlaceGuid\n\n{place}')
            input("Press Enter to continue... (entry will be skipped)")
            return None, post_geoaccuracy, None
        lon_center = 0
        lat_center = 0
        bounding_box = place.get('bounding_box')
        if bounding_box:
            bound_coordinates = bounding_box.get('coordinates')
            if bound_coordinates:
                bounding_box_points = bound_coordinates[0]
            lim_y_min, lim_y_max, lim_x_min, lim_x_max = \
                HF.get_rectangle_bounds(bounding_box_points)
            bound_points_shapely = \
                geometry.MultiPoint([(lim_x_min, lim_y_min),
                                     (lim_x_max, lim_y_max)])
            # True centroid (coords may be multipoint)
            lon_center = bound_points_shapely.centroid.coords[0][0]
            lat_center = bound_points_shapely.centroid.coords[0][1]
        place_type = place.get('place_type')
        if place_type == "country":
            # country_guid
            # in case of country,
            # we do not need to save the GUID from Twitter
            # - country_code is already unique
            country_code = place.get('country_code')
            if country_code:
                place_record = HF.new_lbsn_record_with_id(
                    lbsn.Country(), place.get('country_code'),
                    self.origin)
                if not post_geoaccuracy:
                    post_geoaccuracy = lbsn.Post.COUNTRY
            else:
                self.log.warning(
                    f'No country_code\n\n{place}. '
                    f'PlaceEntry will be skipped..')
                return None, post_geoaccuracy, None
        elif place_type in ("city", "neighborhood", "admin"):
            # city_guid
            place_record = HF.new_lbsn_record_with_id(
                lbsn.City(), place.get('id'), self.origin)
            if not place_type == "city":
                place_record.sub_type = place_type
            if not post_geoaccuracy or post_geoaccuracy == lbsn.Post.COUNTRY:
                post_geoaccuracy = lbsn.Post.CITY
        elif place_type == "poi":
            # place_guid
            # For POIs, lbsn.City is not available on Twitter
            place_record = HF.new_lbsn_record_with_id(lbsn.Place(),
                                                      place.get(
                'id'),
                self.origin)
            if not post_geoaccuracy or post_geoaccuracy in (
                    lbsn.Post.COUNTRY, lbsn.Post.CITY):
                post_geoaccuracy = lbsn.Post.PLACE
        else:
            self.log.warning(f'No lbsn.Place Type Detected: {place}')
        # for some reason, twitter place entities sometimes contain
        # linebreaks or whitespaces. We don't want this.
        place_name = place.get('name').replace('\n\r', '')
        # remove multiple whitespace
        place_name = re.sub(' +', ' ', place_name)
        if place_type == "poi" or \
           user_language is None \
           or not user_language.language_short \
           or user_language.language_short in ('en', 'und'):
            # At the moment, English name are the main references;
            # all other language specific references are stored in
            # name_alternatives - except for places, where twitter
            # has no alternative place names
            # Bugfix necessary: some English names get still saved
            # as name_alternatives
            place_record.name = place_name
        else:
            place_record.name_alternatives.append(place_name)
        place_record.url = place.get('url')
        place_record.geom_center = "POINT(%s %s)" % (lon_center, lat_center)
        if bounding_box and bound_coordinates:
            # prints: 'POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0))'
            place_record.geom_area = Polygon(bounding_box_points).wkt
        ref_country_record = None
        if not isinstance(place_record, lbsn.Country):
            ref_country_code = place.get('country_code')
            if ref_country_code:
                ref_country_record = \
                    HF.new_lbsn_record_with_id(lbsn.Country(),
                                               ref_country_code,
                                               self.origin)
                # At the moment, only English name references are processed
                if user_language is None \
                   or not user_language.language_short \
                   or user_language.language_short in ('en', 'und'):
                    ref_country_record.name = place.get(
                        'country')  # Needs to be saved
                else:
                    alt_name = place.get('country')
                    ref_country_record.name_alternatives.append(alt_name)
                self.lbsn_records.append(ref_country_record)
        if post_geoaccuracy == lbsn.Post.CITY and ref_country_record:
            # country_pkey only on lbsn.City(), lbsn.Place() has city_pkey,
            # but this is not available for Twitter
            place_record.country_pkey.CopyFrom(ref_country_record.pkey)
        # log.debug(f'Final lbsn.Place Record: {placeRecord}')
        return place_record, post_geoaccuracy, ref_country_record



# -*- coding: utf-8 -*-

"""
Shared structure and mapping between DB and Proto LBSN Structure.
"""

# pylint: disable=no-member


from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.tools.helper_functions import HelperFunctions as HF


class ProtoLBSNMapping():
    """ Methods to map ProtoBuf structure to PG SQL structure."""

    @staticmethod
    def get_header_for_type(desc_name):
        """ Will return lbsn-db headers for proto buf record types

        - in correct order
        - ready for copy-from pg import
        """
        dict_switcher = \
            {lbsn.Origin.DESCRIPTOR.name:
             'origin_id, name',
             lbsn.City.DESCRIPTOR.name:
             'origin_id, city_guid, name, name_alternatives, geom_center, '
             'geom_area, url, country_guid, sub_type',
             lbsn.Country.DESCRIPTOR.name:
             'origin_id, country_guid, name, name_alternatives, geom_center, '
             'geom_area, url',
             lbsn.Place.DESCRIPTOR.name:
             'origin_id, place_guid, name, name_alternatives, geom_center, '
             'geom_area, url, city_guid, post_count, place_description, '
             'place_website, place_phone, address, zip_code, attributes, '
             'checkin_count, like_count, parent_places',
             lbsn.Post.DESCRIPTOR.name:
             'origin_id, post_guid, post_latlng, place_guid, city_guid, '
             'country_guid, post_geoaccuracy, user_guid, post_create_date, '
             'post_publish_date, post_body, post_language, user_mentions, '
             'hashtags, emoji, post_like_count, post_comment_count, '
             'post_views_count, post_title, post_thumbnail_url, post_url, '
             'post_type, post_filter, post_quote_count, post_share_count, '
             'input_source, post_content_license',
             lbsn.PostReaction.DESCRIPTOR.name:
             'origin_id, reaction_guid, reaction_latlng, user_guid, '
             'referencedPost_guid, referencedPostreaction_guid, '
             'reaction_type, reaction_date, reaction_content, '
             'reaction_like_count, user_mentions',
             lbsn.User.DESCRIPTOR.name:
             'origin_id, user_guid, user_name, user_fullname, follows, '
             'followed, group_count, biography, post_count, is_private, '
             'url, is_available, user_language, user_location, '
             'user_location_geom, liked_count, active_since, '
             'profile_image_url, user_timezone, user_utc_offset, '
             'user_groups_member, user_groups_follows',
             lbsn.Event.DESCRIPTOR.name:
             'origin_id, event_guid, name, event_latlng, event_area, '
             'event_website, event_date, event_date_start, event_date_end, '
             'duration, place_guid, city_guid, country_guid, user_guid, '
             'event_description, event_type, event_share_count, '
             'event_like_count, event_comment_count, event_views_count, '
             'event_engage_count',
             lbsn.UserGroup.DESCRIPTOR.name:
             'origin_id, usergroup_guid, usergroup_name, '
             'usergroup_description, member_count, '
             'usergroup_createdate, user_owner',
             '_user_mentions_user':
             'origin_id, user_guid, mentioneduser_guid',
             '_user_follows_group':
             'origin_id, user_guid, group_guid',
             '_user_memberof_group':
             'origin_id, user_guid, group_guid',
             '_user_connectsto_user':
             'origin_id, user_guid, connectedto_user_guid',
             '_user_friends_user':
             'origin_id, user_guid, friend_guid'
             }
        return dict_switcher.get(desc_name)

    def func_prepare_selector(self, record):
        """Select correct prepare function according to record type"""
        dict_switcher = {
            lbsn.Origin().DESCRIPTOR.name: self.prepare_lbsn_origin,
            lbsn.Country().DESCRIPTOR.name: self.prepare_lbsn_country,
            lbsn.City().DESCRIPTOR.name: self.prepare_lbsn_city,
            lbsn.Place().DESCRIPTOR.name: self.prepare_lbsn_place,
            lbsn.User().DESCRIPTOR.name: self.prepare_lbsn_user,
            lbsn.UserGroup().DESCRIPTOR.name: self.prepare_lbsn_usergroup,
            lbsn.Post().DESCRIPTOR.name: self.prepare_lbsn_post,
            lbsn.Event().DESCRIPTOR.name: self.prepare_lbsn_event,
            lbsn.PostReaction().DESCRIPTOR.name: self.prepare_lbsn_postreaction,
            lbsn.Relationship().DESCRIPTOR.name: self.prepare_lbsn_relation
        }
        prepare_function = dict_switcher.get(record.DESCRIPTOR.name)
        return prepare_function(record)

    @classmethod
    def prepare_lbsn_country(cls, record):
        """Get common attributes for records of type lbsn.Place"""
        place_record = PlaceBaseAttrShared(record)
        prepared_record = (place_record.origin_id,
                           place_record.guid,
                           place_record.name,
                           place_record.name_alternatives,
                           HF.return_ewkb_from_geotext(
                               place_record.geom_center),
                           HF.return_ewkb_from_geotext(place_record.geom_area),
                           place_record.url)
        return prepared_record

    @classmethod
    def prepare_lbsn_origin(cls, record):
        """Get common attributes for records of type lbsn.Origin"""
        prepared_record = (record.origin_id,
                           record.name)
        return prepared_record

    @classmethod
    def prepare_lbsn_city(cls, record):
        """Get common attributes for records of type lbsn.City"""
        place_record = PlaceBaseAttrShared(record)
        country_guid = HF.null_check(record.country_pkey.id)
        sub_type = HF.null_check(record.sub_type)
        prepared_record = (place_record.origin_id,
                           place_record.guid,
                           place_record.name,
                           place_record.name_alternatives,
                           HF.return_ewkb_from_geotext(
                               place_record.geom_center),
                           HF.return_ewkb_from_geotext(
                               place_record.geom_area),
                           place_record.url,
                           country_guid,
                           sub_type)
        return prepared_record

    @classmethod
    def prepare_lbsn_place(cls, record):
        """Get common attributes for records of type lbsn.Place"""
        place_record = PlaceBaseAttrShared(record)
        # get additional attributes specific to places
        city_guid = HF.null_check(record.city_pkey.id)
        post_count = HF.null_check(record.post_count)
        place_description = HF.null_check(record.place_description)
        place_website = HF.null_check(record.place_website)
        place_phone = HF.null_check(record.place_phone)
        address = HF.null_check(record.address)
        zip_code = HF.null_check(record.zip_code)
        attributes = HF.map_to_dict(HF.null_check(record.attributes))
        checkin_count = HF.null_check(record.checkin_count)
        like_count = HF.null_check(record.like_count)
        parent_places = HF.null_check(record.parent_places)
        prepared_record = (place_record.origin_id,
                           place_record.guid,
                           place_record.name,
                           place_record.name_alternatives,
                           HF.return_ewkb_from_geotext(
                               place_record.geom_center),
                           HF.return_ewkb_from_geotext(
                               place_record.geom_area),
                           place_record.url,
                           city_guid,
                           post_count,
                           place_description,
                           place_website,
                           place_phone,
                           address,
                           zip_code,
                           attributes,
                           checkin_count,
                           like_count,
                           parent_places)
        return prepared_record

    @classmethod
    def prepare_lbsn_user(cls, record):
        """Get common attributes for records of type lbsn.User"""
        user_record = UserAttrShared(record)
        prepared_record = (user_record.origin_id,
                           user_record.guid,
                           user_record.user_name,
                           user_record.user_fullname,
                           user_record.follows,
                           user_record.followed,
                           user_record.group_count,
                           user_record.biography,
                           user_record.post_count,
                           user_record.is_private,
                           user_record.url,
                           user_record.is_available,
                           user_record.user_language,
                           user_record.user_location,
                           HF.return_ewkb_from_geotext(
                               user_record.user_location_geom),
                           user_record.liked_count,
                           user_record.active_since,
                           user_record.profile_image_url,
                           user_record.user_timezone,
                           user_record.user_utc_offset,
                           user_record.user_groups_member,
                           user_record.user_groups_follows)
        return prepared_record

    @classmethod
    def prepare_lbsn_usergroup(cls, record):
        """Get common attributes for records of type Userroup"""
        user_group_record = UsergroupAttrShared(record)
        prepared_record = (user_group_record.origin_id,
                           user_group_record.guid,
                           user_group_record.usergroup_name,
                           user_group_record.usergroup_description,
                           user_group_record.member_count,
                           user_group_record.usergroup_createdate,
                           user_group_record.user_owner)
        return prepared_record

    @classmethod
    def prepare_lbsn_post(cls, record):
        """Get common attributes for records of type lbsn.Post"""
        post_record = PostAttrShared(record)
        prepared_record = (post_record.origin_id,
                           post_record.guid,
                           HF.return_ewkb_from_geotext(
                               post_record.post_latlng),
                           post_record.place_guid,
                           post_record.city_guid,
                           post_record.country_guid,
                           post_record.post_geoaccuracy,
                           post_record.user_guid,
                           post_record.post_create_date,
                           post_record.post_publish_date,
                           post_record.post_body,
                           post_record.post_language,
                           post_record.user_mentions,
                           post_record.hashtags,
                           post_record.emoji,
                           post_record.post_like_count,
                           post_record.post_comment_count,
                           post_record.post_views_count,
                           post_record.post_title,
                           post_record.post_thumbnail_url,
                           post_record.post_url,
                           post_record.post_type,
                           post_record.post_filter,
                           post_record.post_quote_count,
                           post_record.post_share_count,
                           post_record.input_source,
                           post_record.post_content_license)
        return prepared_record

    @classmethod
    def prepare_lbsn_event(cls, record):
        """Get common attributes for records of type lbsn.Post"""
        event_record = EventAttrShared(record)
        prepared_record = (event_record.origin_id,
                           event_record.event_guid,
                           event_record.name,
                           HF.return_ewkb_from_geotext(
                               event_record.event_latlng),
                           HF.return_ewkb_from_geotext(
                               event_record.event_area),
                           event_record.event_website,
                           event_record.event_date,
                           event_record.event_date_start,
                           event_record.event_date_end,
                           event_record.duration,
                           event_record.place_guid,
                           event_record.city_guid,
                           event_record.country_guid,
                           event_record.user_guid,
                           event_record.event_description,
                           event_record.event_type,
                           event_record.event_share_count,
                           event_record.event_like_count,
                           event_record.event_comment_count,
                           event_record.event_views_count,
                           event_record.event_engage_count)
        return prepared_record

    @classmethod
    def prepare_lbsn_postreaction(cls, record):
        """Get common attributes for records of type lbsn.PostReaction"""
        post_reaction_record = PostreactionAttrShared(record)
        prepared_record = (post_reaction_record.origin_id,
                           post_reaction_record.guid,
                           HF.return_ewkb_from_geotext(
                               post_reaction_record.reaction_latlng),
                           post_reaction_record.user_guid,
                           post_reaction_record.referenced_post,
                           post_reaction_record.referenced_postreaction,
                           post_reaction_record.reaction_type,
                           post_reaction_record.reaction_date,
                           post_reaction_record.reaction_content,
                           post_reaction_record.reaction_like_count,
                           post_reaction_record.user_mentions)
        return prepared_record

    @classmethod
    def prepare_lbsn_relation(cls, record):
        """Get common attributes for records of type lbsn Relation"""
        relations_record = RelationAttrShared(record)
        prepared_typerecord_tuple = \
            (relations_record.rel_type,
             cls.prepareRecordValues(relations_record.origin_id,
                                     relations_record.guid,
                                     relations_record.guid_rel))
        return prepared_typerecord_tuple


class PlaceBaseAttrShared():
    """Shared structure for lbsn.Place Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    Note that lbsn.Country, lbsn.City and lbsn.Place have the same structure.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.Place()  # init empty structure
        self.origin_id = record.pkey.origin.origin_id  # = 3
        self.guid = record.pkey.id
        self.name = HF.null_check(record.name)
        # because ProtoBuf Repeated Field does not support distinct rule,
        # we remove any duplicates in list fields prior to submission here
        self.name_alternatives = list(set(record.name_alternatives))
        if self.name and self.name in self.name_alternatives:
            self.name_alternatives.remove(self.name)
        self.url = HF.null_check(record.url)
        self.geom_center = HF.null_check(record.geom_center)
        self.geom_area = HF.null_check(record.geom_area)


class UserAttrShared():
    """Shared structure for lbsn.User Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.User()
        self.origin_id = record.pkey.origin.origin_id
        self.guid = record.pkey.id
        self.user_name = HF.null_check(record.user_name)
        self.user_fullname = HF.null_check(record.user_fullname)
        self.follows = HF.null_check(record.follows)
        self.followed = HF.null_check(record.followed)
        self.group_count = HF.null_check(record.group_count)
        self.biography = HF.null_check(record.biography)
        self.post_count = HF.null_check(record.post_count)
        self.url = HF.null_check(record.url)
        self.is_private = HF.null_check(record.is_private)
        self.is_available = HF.null_check(record.is_available)
        self.user_language = HF.null_check(record.user_language.language_short)
        self.user_location = HF.null_check(record.user_location)
        self.user_location_geom = HF.null_check(record.user_location_geom)
        self.liked_count = HF.null_check(record.liked_count)
        self.active_since = HF.null_check_datetime(record.active_since)
        self.profile_image_url = HF.null_check(record.profile_image_url)
        self.user_timezone = HF.null_check(record.user_timezone)
        self.user_utc_offset = HF.null_check(record.user_utc_offset)
        self.user_groups_member = list(set(record.user_groups_member))
        self.user_groups_follows = list(set(record.user_groups_follows))


class UsergroupAttrShared():
    """Shared structure for Usergroup Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.UserGroup()
        self.origin_id = record.pkey.origin.origin_id
        self.guid = record.pkey.id
        self.usergroup_name = HF.null_check(record.usergroup_name)
        self.usergroup_description = HF.null_check(
            record.usergroup_description)
        self.member_count = HF.null_check(record.member_count)
        self.usergroup_createdate = HF.null_check_datetime(
            record.usergroup_createdate)
        self.user_owner = HF.null_check(record.user_owner_pkey.id)


class PostAttrShared():
    """Shared structure for lbsn.Post Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.Post()

        self.origin_id = record.pkey.origin.origin_id
        self.guid = record.pkey.id
        self.post_latlng = HF.null_geom_check(record.post_latlng)
        self.place_guid = HF.null_check(record.place_pkey.id)
        self.city_guid = HF.null_check(record.city_pkey.id)
        self.country_guid = HF.null_check(record.country_pkey.id)
        self.post_geoaccuracy = HF.turn_lower(HF.null_check(
            lbsn.Post().PostGeoaccuracy.Name(record.post_geoaccuracy)))
        self.user_guid = HF.null_check(record.user_pkey.id)
        self.post_create_date = HF.null_check_datetime(record.post_create_date)
        self.post_publish_date = HF.null_check_datetime(
            record.post_publish_date)
        self.post_body = HF.null_check(record.post_body)
        self.post_language = HF.null_check(record.post_language.language_short)
        self.user_mentions = list(
            set([pkey.id for pkey in record.user_mentions_pkey]))
        self.hashtags = list(set(record.hashtags))
        self.emoji = list(set(record.emoji))
        self.post_like_count = HF.null_check(record.post_like_count)
        self.post_comment_count = HF.null_check(record.post_comment_count)
        self.post_views_count = HF.null_check(record.post_views_count)
        self.post_title = HF.null_check(record.post_title)
        self.post_thumbnail_url = HF.null_check(record.post_thumbnail_url)
        self.post_url = HF.null_check(record.post_url)
        self.post_type = HF.turn_lower(HF.null_check(
            lbsn.Post().PostType.Name(record.post_type)))
        self.post_filter = HF.null_check(record.post_filter)
        self.post_quote_count = HF.null_check(record.post_quote_count)
        self.post_share_count = HF.null_check(record.post_share_count)
        self.input_source = HF.null_check(record.input_source)
        self.post_content_license = HF.null_check(record.post_content_license)
        # optional:
        self.latitude = 0
        self.longitude = 0

class EventAttrShared():
    """Shared structure for lbsn.Event Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.Event()

        self.origin_id = record.pkey.origin.origin_id
        self.event_guid = record.pkey.id
        self.name = HF.null_check(record.name)
        self.event_latlng = HF.null_geom_check(record.event_latlng)
        self.event_area = HF.null_check(record.event_area)
        self.event_website = HF.null_check(record.event_website)
        self.event_date = HF.null_check_datetime(record.event_date)
        self.event_date_start = HF.null_check_datetime(record.event_date_start)
        self.event_date_end = HF.null_check_datetime(record.event_date_end)
        self.duration = HF.null_check(record.duration)
        self.place_guid = HF.null_check(record.place_pkey.id)
        self.city_guid = HF.null_check(record.city_pkey.id)
        self.country_guid = HF.null_check(record.country_pkey.id)
        self.user_guid = HF.null_check(record.user_pkey.id)
        self.event_description = HF.null_check(record.event_description)
        self.event_type = HF.null_check(record.event_type)
        self.event_share_count = HF.null_check(record.event_share_count)
        self.event_like_count = HF.null_check(record.event_like_count)
        self.event_comment_count = HF.null_check(record.event_comment_count)
        self.event_views_count = HF.null_check(record.event_views_count)
        self.event_engage_count = HF.null_check(record.event_engage_count)

class PostreactionAttrShared():
    """Shared structure for Postreaction Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, record=None):
        if record is None:
            record = lbsn.PostReaction()
        self.origin_id = record.pkey.origin.origin_id
        self.guid = record.pkey.id
        self.reaction_latlng = HF.null_geom_check(record.reaction_latlng)
        self.user_guid = HF.null_check(record.user_pkey.id)
        self.referenced_post = HF.null_check(record.referencedPost_pkey.id)
        self.referenced_postreaction = HF.null_check(
            record.referencedPostreaction_pkey.id)
        self.reaction_type = HF.turn_lower(HF.null_check(
            lbsn.PostReaction().ReactionType.Name(record.reaction_type)))
        self.reaction_date = HF.null_check_datetime(record.reaction_date)
        self.reaction_content = HF.null_check(record.reaction_content)
        self.reaction_like_count = HF.null_check(record.reaction_like_count)
        self.user_mentions = list(
            set([pkey.id for pkey in record.user_mentions_pkey]))


class RelationAttrShared():
    """Shared structure for Relation Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    """

    def __init__(self, relationship=None):
        if relationship is None:
            relationship = lbsn.Relationship()
        self.origin_id = relationship.pkey.relation_to.origin.origin_id
        self.guid = relationship.pkey.relation_to.id
        self.guid_rel = relationship.pkey.relation_from.id
        self.rel_type = HF.null_check(lbsn.Relationship().RelationshipType.Name(
            relationship.relationship_type)).lower()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lbsntransform package script to load, format and store data
from and to common lbsn structure

Import options:
    - Postgres Database or
    - local CSV/json/stacked json import
Output options:
    - Postgres Database or
    - local ProtoBuf and CSV Import (prepared for Postgres /Copy)
"""

__author__ = "Alexander Dunkel"
__license__ = "GNU GPLv3"
# version: see version.py

import sys

from lbsntransform.tools.helper_functions import HelperFunctions as HF
from lbsntransform.output.shared_structure import TimeMonitor
from lbsntransform.input.load_data import LoadData
from lbsntransform.config.config import BaseConfig
from lbsntransform.lbsntransform_ import LBSNTransform


def main():
    """ Main function for cli-mode to process data
    from postgres db or local file input
    to postgres db or local file output
    """

    # Load Config, will be overwritten if args are given
    config = BaseConfig()
    # Parse args
    config.parse_args()

    # initialize mapping class
    # depending on lbsn origin
    # e.g. 1 = Instagram,
    #      2 = Flickr, 2.1 = Flickr YFCC100m,
    #      3 = Twitter)
    importer = HF.load_importer_mapping_module(
        config.origin, config.mappings_path)

    # initialize lbsntransform
    lbsntransform = LBSNTransform(
        importer=importer,
        logging_level=config.logging_level,
        is_local_input=config.is_local_input,
        transfer_count=config.transfer_count,
        csv_output=config.csv_output,
        csv_suppress_linebreaks=config.csv_suppress_linebreaks,
        dbuser_output=config.dbuser_output,
        dbserveraddress_output=config.dbserveraddress_output,
        dbname_output=config.dbname_output,
        dbpassword_output=config.dbpassword_output,
        dbserverport_output=config.dbserverport_output,
        dbformat_output=config.dbformat_output,
        dbuser_input=config.dbuser_input,
        dbserveraddress_input=config.dbserveraddress_input,
        dbname_input=config.dbname_input,
        dbpassword_input=config.dbpassword_input,
        dbserverport_input=config.dbserverport_input,
        dbuser_hllworker=config.dbuser_hllworker,
        dbserveraddress_hllworker=config.dbserveraddress_hllworker,
        dbname_hllworker=config.dbname_hllworker,
        dbpassword_hllworker=config.dbpassword_hllworker,
        dbserverport_hllworker=config.dbserverport_hllworker,
        include_lbsn_bases=config.include_lbsn_bases,
        dry_run=config.dry_run,
        hmac_key=config.hmac_key)

    # initialize input reader
    input_data = LoadData(
        importer=importer,
        is_local_input=config.is_local_input,
        startwith_db_rownumber=config.startwith_db_rownumber,
        skip_until_file=config.skip_until_file,
        cursor_input=lbsntransform.cursor_input,
        input_path=config.input_path,
        recursive_load=config.recursive_load,
        local_file_type=config.local_file_type,
        endwith_db_rownumber=config.endwith_db_rownumber,
        is_stacked_json=config.is_stacked_json,
        is_line_separated_json=config.is_line_separated_json,
        csv_delim=config.csv_delim,
        use_csv_dictreader=config.use_csv_dictreader,
        input_lbsn_type=config.input_lbsn_type,
        dbformat_input=config.dbformat_input,
        geocode_locations=config.geocode_locations,
        ignore_input_source_list=config.ignore_input_source_list,
        disable_reactionpost_ref=config.disable_reactionpost_ref,
        map_relations=config.map_relations,
        transfer_reactions=config.transfer_reactions,
        ignore_non_geotagged=config.ignore_non_geotagged,
        min_geoaccuracy=config.min_geoaccuracy,
        source_web=config.source_web,
        skip_until_record=config.skip_until_record,
        zip_records=config.zip_records,
        include_lbsn_objects=config.include_lbsn_objects,
        override_lbsn_query_schema=config.override_lbsn_query_schema)

    # Manually add entries that need submission prior to parsing data
    # add_bundestag_group_example(import_mapper)

    # init time monitoring
    how_long = TimeMonitor()

    # read and process unfiltered input records from csv
    # start settings
    with input_data as records:
        for record in records:
            lbsntransform.add_processed_records(record)
            # report progress
            if lbsntransform.processed_total % 1000 == 0:
                stats_str = HF.report_stats(
                    input_data.count_glob,
                    input_data.continue_number,
                    lbsntransform.lbsn_records)
                print(stats_str, end='\r')
                sys.stdout.flush()
            if (config.transferlimit and
                    lbsntransform.processed_total >= config.transferlimit):
                break

    # finalize output (close db connection, submit remaining)
    lbsntransform.log.info(
        f'\nTransferring remaining '
        f'{lbsntransform.lbsn_records.count_glob} to db.. '
        f'{HF.null_notice(input_data.import_mapper.null_island)})')
    lbsntransform.finalize_output()

    # final report
    lbsntransform.log.info(
        f'\n\n{"".join([f"(Dry Run){chr(10)}" if config.dry_run else ""])}'
        f'Processed {input_data.count_glob} input records '
        f'(Input {input_data.start_number} to '
        f'{input_data.continue_number}). '
        f'\n\nIdentified {lbsntransform.processed_total} LBSN records, '
        f'with {lbsntransform.lbsn_records.count_glob_total} '
        f'distinct LBSN records overall. '
        f'{HF.get_skipped_report(input_data.import_mapper)}. '
        f'Merged {lbsntransform.lbsn_records.count_dup_merge_total} '
        f'duplicate records.')
    lbsntransform.log.info(
        f'\n{HF.get_count_stats(lbsntransform.lbsn_records)}')

    lbsntransform.log.info(f'Done. {how_long.stop_time()}')

    lbsntransform.close_log()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Hll bases for temporal facet
"""

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.hll import hll_bases as hll

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF
from lbsntransform.tools.helper_functions import HelperFunctions as HF

FACET = 'temporal'


class TemporalBase(hll.HllBase):
    """Intermediate temporal base class that extends HllBase
    for a number of temporal classes
    """

    def __init__(self):
        super().__init__()


class TimestampBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='timestamp')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key["timestamp"] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        self.key["timestamp"] = post_date_time


class DateBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='date')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['date'] = None
        self.attrs['name'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        if post_date_time is None:
            return
        # optional: add name of date (e.g. "New Year's Day")
        self.key['date'] = post_date_time.date()


class MonthBase(TemporalBase):
    """Extends Temporal Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='month')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['year'] = None
        self.key['month'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        if post_date_time is None:
            return
        date = post_date_time.date()
        self.key['year'] = date.year
        self.key['month'] = date.month


class YearBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='year')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['year'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        if post_date_time is None:
            return
        date = post_date_time.date()
        self.key['year'] = date.year


class TimeofdayBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='timeofday')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['timeofday'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        # remove microseconds from datetime
        self.key['timeofday'] = post_date_time.time.replace(
            microsecond=0)


class HourofdayBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='hourofday')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['hourofday'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        # remove seconds and microseconds from datetime
        self.key['hourofday'] = post_date_time.time.replace(
            second=0, microsecond=0)


class DayofweekBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='dayofweek')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['weekday'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        self.key['weekday'] = post_date_time.weekday


class DayofmonthBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='dayofmonth')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['dayofmonth'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        self.key['dayofmonth'] = post_date_time.day


class DayofyearBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='dayofyear')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['month'] = None
        self.key['day'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        self.key['month'] = post_date_time.month
        self.key['day'] = post_date_time.day


class MonthofyearBase(TemporalBase):
    """Extends Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='monthofyear')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        self.key['monthofyear'] = None
        if record is None:
            # init empty
            return
        post_date_time = HLF.merge_dates_post(
            record)
        self.key['monthofyear'] = post_date_time.month

class MonthHashtagBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """
    NAME = hll.HllBaseRef(facet=FACET, base='_month_hashtag')

    def __init__(self, record: lbsn.Post = None, hashtag: str = None):
        super().__init__()
        self.key['year'] = None
        self.key['month'] = None
        self.key['hashtag'] = None
        if hashtag is None:
            # init empty
            return
        self.key['hashtag'] = hashtag.lower()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            post_date_time = HLF.merge_dates_post(
                record)
            if post_date_time is None:
                return
            date = post_date_time.date()
            self.key['year'] = date.year
            self.key['month'] = date.month
        else:
            raise ValueError(
                "Parsing of MonthHashtagBase only supported "
                "from lbsn.Post")

class MonthLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """
    NAME = hll.HllBaseRef(facet=FACET, base='_month_latlng')

    def __init__(self, record: lbsn.Post = None, hashtag: str = None):
        super().__init__()
        self.key['year'] = None
        self.key['month'] = None
        self.key['latitude'] = None
        self.key['longitude'] = None
        self.attrs['latlng_geom'] = None
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            post_date_time = HLF.merge_dates_post(
                record)
            if post_date_time is None:
                return
            date = post_date_time.date()
            self.key['year'] = date.year
            self.key['month'] = date.month

            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['latitude'] = coordinates.lat
            self.key['longitude'] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['latlng_geom'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
        else:
            raise ValueError(
                "Parsing of MonthLatLngBase only supported "
                "from lbsn.Post")


# -*- coding: utf-8 -*-

"""
Module for storing common Proto LBSN Structure to PG DB.
"""

# pylint: disable=no-member

import traceback
import logging
import sys
from typing import Any, Dict, List, Tuple, Union

import psycopg2

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.tools.helper_functions import HelperFunctions as HF

from lbsntransform.output.csv.store_csv import LBSNcsv
from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.output.hll.base import social, spatial, temporal, topical
from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF
from lbsntransform.output.hll.shared_structure_proto_hlldb import ProtoHLLMapping
from lbsntransform.output.hll.sql_hll import HLLSql
from lbsntransform.output.lbsn.shared_structure_proto_lbsndb import ProtoLBSNMapping
from lbsntransform.output.lbsn.sql_lbsn import LBSNSql


class LBSNTransfer():
    """Transfer converted lbsn records
    to various output formats (CSV, raw-db, hll-db)
    """

    def __init__(self, db_cursor=None,
                 db_connection=None,
                 commit_volume=None,
                 disable_reaction_post_ref=0,
                 store_csv=None,
                 SUPPRESS_LINEBREAKS=True,
                 dbformat_output="lbsn",
                 hllworker_cursor=None,
                 include_lbsn_bases=None,
                 dry_run: bool = None):
        self.db_cursor = db_cursor
        self.db_connection = db_connection
        self.dry_run = dry_run
        if not self.db_cursor:
            print("CSV Output Mode.")
        self.count_entries_commit = 0
        self.count_entries_store = 0
        self.count_affected = 0
        if store_csv:
            dbformat_output = "lbsn"
        if commit_volume is None:
            # due to more output bases
            # increase commit volume on hll output
            if dbformat_output == "lbsn":
                commit_volume = 10000
            else:
                commit_volume = 100000
        self.commit_volume = commit_volume
        self.store_volume = 500000
        self.count_glob = 0
        self.null_island_count = 0
        self.disable_reaction_post_ref = disable_reaction_post_ref
        self.log = logging.getLogger('__main__')
        self.batched_lbsn_records = {
            lbsn.Origin.DESCRIPTOR.name: list(),
            lbsn.Country.DESCRIPTOR.name: list(),
            lbsn.City.DESCRIPTOR.name: list(),
            lbsn.Place.DESCRIPTOR.name: list(),
            lbsn.User.DESCRIPTOR.name: list(),
            lbsn.UserGroup.DESCRIPTOR.name: list(),
            lbsn.Post.DESCRIPTOR.name: list(),
            lbsn.PostReaction.DESCRIPTOR.name: list(),
            lbsn.Relationship.DESCRIPTOR.name: list()
        }
        # dynamially register base classes from base module
        hll.register_classes()
        # this is the global dict of measures that are currently supported,
        # bases not registered here will not be measured
        self.batched_hll_records = {
            spatial.LatLngBase.NAME: dict(),
            spatial.PlaceBase.NAME: dict(),
            temporal.DateBase.NAME: dict(),
            temporal.MonthBase.NAME: dict(),
            temporal.YearBase.NAME: dict(),
            temporal.MonthLatLngBase.NAME: dict(),
            temporal.MonthHashtagBase.NAME: dict(),
            topical.TermBase.NAME: dict(),
            topical.HashtagBase.NAME: dict(),
            topical.EmojiBase.NAME: dict(),
            topical.TermLatLngBase.NAME: dict(),
            topical.HashtagLatLngBase.NAME: dict(),
            topical.EmojiLatLngBase.NAME: dict(),
            social.CommunityBase.NAME: dict()
        }
        self.count_round = 0
        if dbformat_output == "lbsn":
            batch_db_volume = 100
        else:
            # increase for less complex but
            # higher quantity of hll records
            batch_db_volume = 20000
        # Records are batched and submitted in
        # one insert with x number of records
        self.batch_db_volume = batch_db_volume
        self.store_csv = store_csv
        self.headers_written = set()
        # self.CSVsuppressLinebreaks = CSVsuppressLinebreaks
        self.dbformat_output = dbformat_output
        if self.dbformat_output == 'lbsn':
            self.db_mapping = ProtoLBSNMapping()
        else:

            if include_lbsn_bases is None:
                include_lbsn_bases = []
            self.db_mapping = ProtoHLLMapping(
                include_lbsn_bases=include_lbsn_bases)
            self.hllworker_cursor = hllworker_cursor

        if self.store_csv:
            self.csv_output = LBSNcsv(SUPPRESS_LINEBREAKS)

    def commit_changes(self):
        """Commit Changes to DB"""
        if self.db_cursor:
            self.db_connection.commit()  #
            self.count_entries_commit = 0

    def store_changes(self):
        """Write changes to CSV"""
        if self.store_csv:
            self.csv_output.clean_csv_batches(
                self.batched_lbsn_records, self.dry_run)
            self.count_entries_store = 0

    def store_origin(self, origin_id, name):
        """Store origin of input source sql"""
        if self.dry_run:
            return
        if self.store_csv:
            origin = lbsn.Origin()
            origin.origin_id = origin_id
            self.csv_output.store_append_batch_to_csv(
                [origin], 0, lbsn.Origin.DESCRIPTOR.name)
            return
        insert_sql = \
            f'''
            INSERT INTO social."origin" (
                origin_id, name)
            VALUES ({origin_id},'{name}')
            ON CONFLICT (origin_id)
            DO NOTHING
            '''
        self.db_cursor.execute(insert_sql)

    def store_lbsn_record_dicts(
            self, lbsn_record_dicts):
        """Main loop for storing lbsn records to CSV or DB

        Arguments:
            field_mapping {field mapping class} -- Import Field mapping class
            with attached data

        order is important here, as PostGres will reject any
        records where Foreign Keys are violated
        therefore, records are processed starting from lowest
        granularity. Order is stored in all_dicts()
        """

        self.count_round += 1
        # self.headersWritten.clear()
        r_cnt = 0
        self.count_affected = 0
        g_cnt = lbsn_record_dicts.get_current_count()
        # clear line
        sys.stdout.write("\033[K")
        for record, type_name in lbsn_record_dicts.get_all_records():
            r_cnt += 1
            print(f'Converting {r_cnt} of {g_cnt} '
                  f'lbsn records ({type_name})..', end='\r')
            self.prepare_lbsn_record(record, type_name)
            self.count_glob += 1  # self.dbCursor.rowcount
            self.count_entries_commit += 1  # self.dbCursor.rowcount
            self.count_entries_store += 1
            if self.db_cursor and (self.count_glob == 100 or
                                   self.count_entries_commit >
                                   self.commit_volume):
                self.commit_changes()
            if self.store_csv and (self.count_entries_store >
                                   self.store_volume):
                self.store_changes()
        # submit remaining rest
        self.submit_all_batches()
        # self.count_affected += x # monitoring
        print(f'\nRound {self.count_round:03d}: '
              f'Updated/Inserted {self.count_glob} records.')

    def prepare_lbsn_record(self, record, record_type):
        """Prepare batched records for submit to either LBSN or HLL db"""
        # clean duplicates in repeated Fields and Sort List
        self.sort_clean_proto_repeated_field(record)
        # store cleaned ProtoBuf records
        # LBSN or HLL output
        if self.dbformat_output == 'lbsn':
            self.batched_lbsn_records[record_type].append(
                record)
        else:
            # extract hll bases and metric from records
            hll_base_metrics = self.db_mapping.extract_hll_base_metrics(
                record, record_type)
            if hll_base_metrics is None:
                # no base metrics extracted
                return
            # update hll dicts
            self.db_mapping.update_hll_dicts(
                self.batched_hll_records, hll_base_metrics)
        # check batched records (and submit)
        self.check_batchvolume_submit()

    def check_batchvolume_submit(self):
        """If any dict contains more values than self.batch_db_volume,
           submit/store all
        """
        if self.dbformat_output == 'lbsn':
            batch_lists = self.batched_lbsn_records.values()
        else:
            batch_lists = self.batched_hll_records.values()
        for batch_list in batch_lists:
            if len(batch_list) >= self.batch_db_volume:
                self.submit_all_batches()

    def submit_all_batches(self):
        """Hook to submit either lbsn or hll records"""
        if self.dbformat_output == 'lbsn':
            self.submit_batches(self.batched_lbsn_records)
        else:
            self.submit_batches(self.batched_hll_records)

    def submit_batches(self, batched_records: Union[
            Dict[str, List[str]],
            Dict[Tuple[str, str], Dict[str, Any]]]):
        """Prepare values for each batch, format sql and submit to db"""
        for record_type, batch_item in batched_records.items():
            if batch_item:
                # if self.storeCSV and not record_type in self.headersWritten:
                #    self.writeCSVHeader(record_type)
                #    self.headersWritten.add(record_type)
                func_select = self.get_prepare_records_func(
                    self.dbformat_output)
                prepared_records = func_select(batch_item)
                self.submit_records(record_type, prepared_records)
                batch_item.clear()

    def get_prepare_records_func(self, dbformat_outpur: str):
        """Selector function to get prepared records (hll/lbsn)
        """
        if dbformat_outpur == 'lbsn':
            return self.get_prepared_lbsn_records
        # hll
        return self.get_prepared_hll_records

    def get_prepared_lbsn_records(
            self, batch_item: List[Any]):
        """Turns proprietary lbsn classes into prepared sql value tuples

        For hll output, this includes calculation of
        shards from individual items using the hll_worker
        """
        prepared_records = []
        for record_item in batch_item:
            try:
                prepared_record = self.db_mapping.func_prepare_selector(
                    record_item)
            except Exception:
                print(
                    f'Could not process record:\n'
                    f'{record_item}')
                track = traceback.format_exc()
                print(f'Traceback: {track}')
                sys.exit(1)
            if prepared_record:
                prepared_records.append(prepared_record)
        return prepared_records

    def prepare_hmac(self, hmac_key: str = None):
        """Update the session hmac key that is used to apply cryptographic hashing

        Defaults to empty string, if no --hmac_key is provided and crypt.salt
        is not set in worker db.
        """
        if hmac_key is None:
            logging.getLogger('__main__').warn(
                "Use of empty hmac_key: For production use, please "
                "make sure that either `crypt.salt` is set on your worker db, "
                "or provide the --hmac_key.")
            self.hllworker_cursor.execute(
            """
            SELECT set_config(
                'crypt.salt',
                COALESCE(current_setting('crypt.salt', 't'), ''),
                false);
            """)
        else:
            self.hllworker_cursor.execute(
                "SET crypt.salt = %s", hmac_key)
        hmac_func_sql = HLLSql.get_hmac_hash_sql()
        self.hllworker_cursor.execute(hmac_func_sql)


    def get_prepared_hll_records(
            self, batch_item: Dict[str, Any]):
        """Turns propietary hll classes into prepared sql value tuples

        This includes calculation of shards from individual items
        using the hll_worker
        """
        hll_items = []  # (base_key, metric_key, item)
        hll_base_records = []  # (base_key, attr1, attr2)
        # the following iteration will
        # loop keys in case of dict
        # and values in case of list
        for index, record_item in enumerate(batch_item.values()):
            # get base record and value
            base = record_item.get_prepared_record()
            if base.record:
                hll_base_records.append(base.record)
                base_metric_item_tuples = HLF.concat_base_metric_item(
                    index, base.metrics)
                # format tuple-values as sql-escaped strings
                value_str = [
                    self.prepare_sqlescaped_values(record) for
                    record in base_metric_item_tuples]
                # add to global list of items to be upserted
                hll_items.extend(value_str)
        # format sql for shard generation
        # get sql escaped values list
        values_str = HF.concat_values_str(hll_items)
        # clear line
        sys.stdout.write("\033[K")
        print(
            f'Calculating hll shards for {len(values_str)} values..',
            end='\r')
        # calculate hll shards from raw values
        hll_shards = HLF.calculate_item_shards(
            self.hllworker_cursor, values_str)
        prepared_records = HLF.concat_base_shards(
            hll_base_records, hll_shards)
        return prepared_records

    def submit_records(
            self, record_type: Union[Tuple[str, str], str],
            prepared_records: List[Tuple[Any]]):
        """Submit/save prepared records to db or csv"""
        if self.store_csv:
            if self.dry_run:
                return
            self.csv_output.store_append_batch_to_csv(
                self.batched_lbsn_records[record_type],
                self.count_round, record_type)
        if self.db_cursor:
            # get sql escaped values list
            sql_escaped_values_list = [
                self.prepare_sqlescaped_values(record) for
                record in prepared_records]
            # concat to single sql str
            values_str = HF.concat_values_str(
                sql_escaped_values_list
            )
            insert_sql = self.insert_sql_selector(values_str, record_type)
            # clear line
            sys.stdout.write("\033[K")
            print(f'Submitting {len(prepared_records)}..', end='\r')
            self.submit_batch(insert_sql)

    def insert_sql_selector(
            self, values_str: str, record_type):
        """ Select function to prepare SQL insert.

        Attributes:
            record_type     type of record
            values_str      values to be inserted
        """
        if self.dbformat_output == 'lbsn':
            sql_selector = LBSNSql.type_sql_mapper()
            prepare_function = sql_selector.get(record_type)
        else:
            # hll
            prepare_function = HLLSql.hll_insertsql
        return prepare_function(values_str, record_type)

    def submit_lbsn_relationships(self):
        """submit relationships of different types

        record[1] is the PostgresQL formatted list of values,
        record[0] is the type of relationship that determines
            the table selection
        """
        select_friends = [relationship[1] for relationship in
                          self.batched_lbsn_records[lbsn.Relationship(
                          ).DESCRIPTOR.name] if relationship[0] == "isfriend"]
        if select_friends:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    select_friends, self.count_round, '_user_friends_user')
            if self.db_cursor:
                args_isfriend = ','.join(select_friends)
                insert_sql = \
                    f'''
                    INSERT INTO interlinkage."_user_friends_user" (
                        {self.typeNamesHeaderDict["_user_friends_user"]})
                    VALUES {args_isfriend}
                    ON CONFLICT (origin_id, user_guid, friend_guid)
                    DO NOTHING
                    '''
                self.submit_batch(insert_sql)
        select_connected = [relationship[1] for relationship in
                            self.batched_lbsn_records[lbsn.Relationship(
                            ).DESCRIPTOR.name] if
                            relationship[0] == "isconnected"]
        if select_connected:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    select_connected, self.count_round, '_user_connectsto_user')
            if self.db_cursor:
                args_isconnected = ','.join(select_connected)
                insert_sql = \
                    f'''
                        INSERT INTO interlinkage."_user_connectsto_user" (
                            {self.typeNamesHeaderDict["_user_connectsto_user"]})
                        VALUES {args_isconnected}
                        ON CONFLICT (origin_id, user_guid,
                            connectedto_user_guid)
                        DO NOTHING
                    '''
                self.submit_batch(insert_sql)
        select_usergroupmember = [relationship[1] for relationship in
                                  self.batched_lbsn_records[lbsn.Relationship(
                                  ).DESCRIPTOR.name] if
                                  relationship[0] == "ingroup"]
        if select_usergroupmember:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    select_usergroupmember, self.count_round,
                    '_user_memberof_group')
            if self.db_cursor:
                args_isingroup = ','.join(select_usergroupmember)
                insert_sql = \
                    f'''
                    INSERT INTO interlinkage."_user_memberof_group" (
                        {self.typeNamesHeaderDict["_user_memberof_group"]})
                    VALUES {args_isingroup}
                    ON CONFLICT (origin_id, user_guid, group_guid)
                    DO NOTHING
                    '''
                self.submit_batch(insert_sql)
        select_usergroupmember = [relationship[1] for relationship in
                                  self.batched_lbsn_records[lbsn.Relationship(
                                  ).DESCRIPTOR.name] if
                                  relationship[0] == "followsgroup"]
        if select_usergroupmember:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    select_usergroupmember, self.count_round,
                    '_user_follows_group')
            if self.db_cursor:
                args_isingroup = ','.join(select_usergroupmember)
                insert_sql = \
                    f'''
                    INSERT INTO interlinkage."_user_follows_group" (
                        {self.typeNamesHeaderDict["_user_follows_group"]})
                    VALUES {args_isingroup}
                    ON CONFLICT (origin_id, user_guid, group_guid)
                    DO NOTHING
                    '''
                self.submit_batch(insert_sql)
        select_usermentions = [relationship[1] for relationship in
                               self.batched_lbsn_records[lbsn.Relationship(
                               ).DESCRIPTOR.name] if
                               relationship[0] == "mentions_user"]
        if select_usermentions:
            if self.store_csv:
                self.csv_output.store_append_batch_to_csv(
                    select_usermentions, self.count_round,
                    '_user_mentions_user')
            if self.db_cursor:
                args_isingroup = ','.join(select_usermentions)
                insert_sql = \
                    f'''
                    INSERT INTO interlinkage."_user_mentions_user" (
                        {self.typeNamesHeaderDict["_user_mentions_user"]})
                    VALUES {args_isingroup}
                    ON CONFLICT (origin_id, user_guid, mentioneduser_guid)
                    DO NOTHING
                    '''
                self.submit_batch(insert_sql)

    def submit_batch(self, insert_sql):
        """Submit Batch to PG DB.

        Needs testing: is using Savepoint for each insert slower
        than rolling back entire commit?
        for performance, see https://stackoverflow.com/questions/
            12206600/how-to-speed-up-insertion-performance-in-postgresql
        or this: https://stackoverflow.com/questions/8134602/
        psycopg2-insert-multiple-rows-with-one-query
        """
        tsuccessful = False
        if self.dry_run:
            return
        self.db_cursor.execute("SAVEPOINT submit_recordBatch")
        while not tsuccessful:
            try:
                self.db_cursor.execute(insert_sql)
            except psycopg2.IntegrityError as einteg:
                if '(post_language)' in einteg.diag.message_detail or \
                        '(user_language)' in einteg.diag.message_detail:
                    # If language does not exist, we'll trust Twitter
                    # and add this to our language list
                    missingLanguage = einteg.diag.message_detail.partition(
                        "language)=(")[2].partition(") is not present")[0]
                    print(
                        f'TransactionIntegrityError, inserting language "'
                        f'{missingLanguage}" first..               ')
                    # self.db_cursor.rollback()
                    self.db_cursor.execute(
                        "ROLLBACK TO SAVEPOINT submit_recordBatch")
                    insert_language_sql = '''
                           INSERT INTO social."language"
                            (language_short, language_name, language_name_de)
                           VALUES (%s,NULL,NULL);
                           '''
                    # submit sql to db
                    self.db_cursor.execute(
                        insert_language_sql, (missingLanguage,))
                    # commit changes so they're available when
                    # try is executed again
                    self.commit_changes()
                    # recreate SAVEPOINT after language insert
                    self.db_cursor.execute("SAVEPOINT submit_recordBatch")
                else:
                    sys.exit(f'{einteg}')
            except psycopg2.DataError as edata:
                sys.exit(f'{edata}\nINSERT SQL WAS: {insert_sql}')
            except ValueError as evalue:
                self.log.warning(f'{evalue}')
                input("Press Enter to continue... (entry will be skipped)")
                self.log.warning(f'{insert_sql}')
                input("args:... ")
                # self.log.warning(f'{args_str}')
                self.db_cursor.execute(
                    "ROLLBACK TO SAVEPOINT submit_recordBatch")
                tsuccessful = True
            except psycopg2.ProgrammingError as epsyco:
                file = open("hll_exc.txt", "w")
                file.write(f'{epsyco}\nINSERT SQL WAS: {insert_sql}')
                file.close()  # This close() is important
                sys.exit(f'{epsyco}\nINSERT SQL WAS: {insert_sql}')
            except psycopg2.errors.DiskFull:
                input("Disk space full. Clean files and continue..")
            else:
                # executed if the try clause does not raise an exception
                self.db_cursor.execute("RELEASE SAVEPOINT submit_recordBatch")
                tsuccessful = True

    def prepare_sqlescaped_values(self, *args):
        """dynamically construct sql value injection

        e.g. record_sql = '''(%s,%s,%s,%s,%s,%s,%s)'''
        """
        record_sql = f'''{','.join('%s' for x in range(0, len(args)))}'''
        # inject values
        prepared_sql_record = self.db_cursor.mogrify(record_sql, tuple(args))
        # mogrify returns a byte object,
        # we decode it so it can be used as a string again
        prepared_sql_record = prepared_sql_record.decode()
        return prepared_sql_record

    @classmethod
    def sort_clean_proto_repeated_field(cls, record):
        """Remove duplicate values in repeated field, sort alphabetically

        ProtocolBuffers has no unique list field type. This function will
        remove duplicates, which is needed for unique compare.

        There is a 'bug' in Python implementation of ProtocolBuffers:
        - depending on the implementation type in use, it is possible
        to spot either 'RepeatedCompositeFieldContainer'
            or 'RepeatedCompositeContainer'
        - solution here: import and compare to both types
        - this is not ideal, since both types are internal to PB and
            subject to change
        - see [proto-bug](https://github.com/protocolbuffers/
            protobuf/issues/3870)
        """
        for descriptor in record.DESCRIPTOR.fields:
            if descriptor.label == descriptor.LABEL_REPEATED:
                x_attr = getattr(record, descriptor.name)
                if x_attr and not HF.is_composite_field_container(
                        x_attr):
                    x_attr_cleaned = set(x_attr)
                    x_attr_sorted = sorted(x_attr_cleaned)
                    # Complete clear of repeated field
                    for _ in range(0, len(x_attr)):
                        x_attr.pop()
                    # add sorted list
                    x_attr.extend(x_attr_sorted)

    def finalize(self):
        """ Final procedure calls:
            - clean and merge csv batches
        """
        if self.store_csv:
            self.csv_output.clean_csv_batches(
                self.batched_lbsn_records, self.dry_run)



# -*- coding: utf-8 -*-

"""
Module for loding data from different sources (CSV, DB, JSON etc.).
"""

import codecs
import csv
import os
import sys
import logging
import traceback
from contextlib import closing
from itertools import zip_longest
from typing import Tuple, List, Union, Iterator, Optional, IO
import psycopg2

import ntpath
import requests
from lbsnstructure import lbsnstructure_pb2 as lbsn

from lbsntransform.tools.db_connection import DBConnection
from lbsntransform.output.shared_structure import GeocodeLocations
from lbsntransform.tools.helper_functions import HelperFunctions as HF
from lbsntransform.input.mappings.db_query import (
    InputSQL, LBSN_SCHEMA, optional_schema_override)

# type alias
LBSNObjects = Union[
    lbsn.CompositeKey, lbsn.Language, lbsn.RelationshipKey, lbsn.City,
    lbsn.Country, lbsn.Origin, lbsn.Place, lbsn.Post,
    lbsn.PostReaction, lbsn.Relationship, lbsn.User,
    lbsn.UserGroup]


class LoadData():
    """
    Class for loding data from different sources (CSV, DB, JSON etc.).
    """

    def __init__(
            self, importer=None, is_local_input=None,
            startwith_db_rownumber=None,
            skip_until_file=None, cursor_input=None, input_path=None,
            recursive_load=None, local_file_type=None,
            endwith_db_rownumber=None, is_stacked_json=None,
            is_line_separated_json=None, csv_delim=None,
            input_lbsn_type=None, dbformat_input=None, geocode_locations=None,
            ignore_input_source_list=None, disable_reactionpost_ref=None,
            map_relations=None, transfer_reactions=None,
            ignore_non_geotagged=None, min_geoaccuracy=None, source_web=None,
            zip_records=None, skip_until_record=None,
            include_lbsn_objects=None, override_lbsn_query_schema=None,
            use_csv_dictreader=None):
        self.is_local_input = is_local_input
        self.start_number = 1
        self.continue_number = None
        self.skip_until_record = None
        if not self.is_local_input:
            # Start Value, Modify to continue from last processing
            self.continue_number = startwith_db_rownumber
            self.start_number = startwith_db_rownumber
        else:
            self.continue_number = 0
            if skip_until_file is not None:
                self.continue_number = skip_until_file
            self.skip_until_record = skip_until_record
        self.source_web = source_web
        if zip_records is None:
            zip_records = False
        self.zip_records = zip_records
        self.cursor_input = cursor_input
        if self.is_local_input and not self.source_web:
            self.filelist = LoadData._read_local_files(
                input_path=input_path, recursive_load=recursive_load,
                local_file_type=local_file_type,
                skip_until_file=skip_until_file)
        elif self.is_local_input and self.source_web:
            self.filelist = input_path
        self.finished = False
        self.dbformat_input = dbformat_input

        self.db_row_number = 0
        self.endwith_db_rownumber = endwith_db_rownumber
        self.is_stacked_json = is_stacked_json
        self.is_line_separated_json = is_line_separated_json
        self.local_file_type = local_file_type
        self.csv_delim = csv_delim
        self.use_csv_dictreader = use_csv_dictreader
        self.file_format = local_file_type
        self.input_lbsn_type = input_lbsn_type
        self.start_id = None
        if include_lbsn_objects is None:
            include_lbsn_objects = ["post"]
        self.include_lbsn_objects = include_lbsn_objects
        self.count_glob = 0
        self.current_source = None
        # self.transferlimit = cfg.transferlimit
        # Optional Geocoding
        self.geocode_dict = None
        if geocode_locations:
            self.geocode_dict = LoadData.load_geocodes(
                geocode_locations)
        # Optional ignore input sources
        self.ignore_sources_set = None
        if ignore_input_source_list:
            self.ignore_sources_set = LoadData.load_ignore_sources(
                ignore_input_source_list)

        self.lbsn_schema = LBSN_SCHEMA
        if override_lbsn_query_schema:
            self.lbsn_schema = optional_schema_override(
                LBSN_SCHEMA, override_lbsn_query_schema)
        # initialize field mapping structure
        self.import_mapper = importer(
            disable_reactionpost_ref,
            self.geocode_dict,
            map_relations,
            transfer_reactions,
            ignore_non_geotagged,
            self.ignore_sources_set,
            min_geoaccuracy)
        self.finished = False

    def __enter__(self) -> Iterator[LBSNObjects]:
        """Main pipeline for reading input data

        Combine multiple generators to single pipeline,
        returned for being processed by with-statement
        """
        if self.cursor_input or self.source_web:
            return self.convert_records(self._process_input())
        else:
            return self.convert_records(
                self._process_input(self._open_input_files()))

    def __exit__(self, exception_type, exception_value, tb_value):
        """Contextmanager exit: nothing to do here if no exception is raised"""
        if any(v is not None for v in [
                exception_type, exception_value, tb_value]):
            # only if any of these variables is not None
            # catch exception and output additional information
            logging.getLogger('__main__').warning(
                f"\nError while reading records: "
                f"{exception_type}\n{exception_value}\n"
                f"{traceback.print_tb(tb_value)}\n")
            logging.getLogger('__main__').warning(
                f"Current source: \n {self.current_source}\n")
            stats_str = HF.report_stats(
                self.count_glob,
                self.continue_number)
            logging.getLogger('__main__').warning(stats_str)
        return False

    def _open_input_files(self) -> Iterator[IO[str]]:
        """Loops input filelist and returns opened file handles"""
        # process localfiles
        for file_name in self.filelist:
            self.continue_number += 1
            self.current_source = file_name
            HF.log_main_debug(
                f'Current file: {ntpath.basename(file_name)}')
            yield open(file_name, 'r', encoding="utf-8", errors='replace')

    def _process_input(
            self,
            file_handles: Iterator[IO[str]] = None) -> Iterator[
                Optional[Tuple[List[str], Optional[str]]]]:
        """File parse for CSV or JSON from open file handle

        Output Generator of type Tuple (1, 2):
            1) a list of post that can be parsed
            2) Optional information regarding type of records in list
        """
        if self.source_web:
            if len(self.filelist) == 1:
                # single web file query
                url = self.filelist[0]
                with closing(requests.get(url, stream=True)) as file_handle:
                    record_reader = csv.reader(
                        codecs.iterdecode(
                            file_handle.iter_lines(), 'utf-8'),  # pylint: disable=maybe-no-member
                        delimiter=self.csv_delim,
                        quotechar='"', quoting=csv.QUOTE_NONE)
                    for record in record_reader:
                        # for record in zip_longest(r1, r2)
                        yield record, None
            else:
                # multiple web file query
                if self.zip_records and len(self.filelist) == 2:
                    # zip 2 web csv sources in parallel, e.g.
                    # zip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
                    url1 = self.filelist[0]
                    url2 = self.filelist[1]
                    with closing(requests.get(url1, stream=True)) as fhandle1, \
                            closing(requests.get(url2, stream=True)) as fhandle2:
                        reader1 = csv.reader(
                            codecs.iterdecode(fhandle1.iter_lines(), 'utf-8'),
                            delimiter=self.csv_delim,
                            quotechar='"', quoting=csv.QUOTE_NONE)
                        reader2 = csv.reader(
                            codecs.iterdecode(fhandle2.iter_lines(), 'utf-8'),
                            delimiter=self.csv_delim,
                            quotechar='"', quoting=csv.QUOTE_NONE)
                        for zipped_record in zip_longest(reader1, reader2):
                            # catch any type-error None record
                            # if zipped_record[0] is None:
                            #     yield zipped_record[1]
                            # if zipped_record[1] is None:
                            #     yield zipped_record[0]
                            # two combine lists
                            try:
                                yield zipped_record[0] + zipped_record[1], None
                            except TypeError:
                                sys.exit(f"Stream appears to have broken. "
                                         f"Check connection and continue at "
                                         f"{self.count_glob}")
                    return
                else:
                    raise ValueError(
                        "Iteration of multiple web sources "
                        "is currently not supported.")
        elif self.is_local_input:
            if file_handles is None:
                raise ValueError("Input is empty")
            if self.zip_records:
                if len(self.filelist) == 2:
                    filehandle1 = next(file_handles)
                    filehandle2 = next(file_handles)
                    for zipped_record in zip_longest(
                            self.fetch_record_from_file(filehandle1),
                            self.fetch_record_from_file(filehandle2)):
                        yield zipped_record[0] + zipped_record[1], None
                else:
                    raise ValueError("Zipping only supported for 2 files.")
            else:
                # local file loop
                for file_handle in file_handles:
                    if not self.file_format == 'json':
                        # csv or txt
                        for record in self.fetch_record_from_file(
                                file_handle):
                            if record:
                                yield record, None
                            else:
                                continue
                    else:
                        # json
                        for record in self.fetch_json_data_from_file(
                                file_handle):
                            if record:
                                yield record, None
                            else:
                                continue
        else:
            # db query
            if self.dbformat_input == "lbsn":
                for lbsn_type, schema_name, table_name, key_col in self.lbsn_schema:
                    if lbsn_type.lower() not in self.include_lbsn_objects:
                        continue
                    while self.cursor_input:
                        records = self.fetch_json_data_from_lbsn(
                            cursor=self.cursor_input,
                            start_id=self.continue_number,
                            schema_name=schema_name, table_name=table_name,
                            key_col=key_col)
                        if records is None:
                            break
                        for record in records:
                            yield record, lbsn_type
                    # reset start cursor
                    # note: this will disable --startwith_db_rownumber for
                    # any further lbsnobjects
                    self.continue_number = None
            elif self.dbformat_input == "json":
                while self.cursor_input:
                    records = self.fetch_json_data_from_lbsn(
                        cursor=self.cursor_input, start_id=self.continue_number)
                    for record in records:
                        yield record, self.input_lbsn_type

    def convert_records(
        self, records: Iterator[Optional[Tuple[List[str], Optional[str]]]]
    ) -> Iterator[List[LBSNObjects]]:
        """Loops input json or csv records,
        converts to ProtoBuf structure and adds to records_dict

        Returns statistic-counts, modifies (adds results to) import_mapper
        """
        for record in records:
            self.count_glob += 1
            # skip records based on count
            if self.skip_until_record and self.skip_until_record > self.count_glob:
                print(f'Skipping record {self.count_glob}', end='\r')
                continue
            record_type = None
            if self.is_local_input or self.dbformat_input == "lbsn":
                single_record = record[0]
                record_type = record[1]
            else:
                # e.g. dbformat_input == "json"
                self.db_row_number = record[0]
                single_record = record[2]
            if LoadData.skip_empty_or_other(single_record):
                # skip empty or malformed records
                continue
            if self.local_file_type == 'json' or not self.is_local_input:
                # note: db-records always returned as json-dict
                lbsn_records = self.import_mapper.parse_json_record(
                    single_record, record_type)
            elif self.local_file_type in ('txt', 'csv'):
                lbsn_records = self.import_mapper.parse_csv_record(
                    single_record, record_type)
            else:
                sys.exit(f'Format {self.local_file_type} not supportet.')
            # return record as pipe
            if lbsn_records is None:
                continue
            for lbsn_record in lbsn_records:
                yield lbsn_record

    @staticmethod
    def skip_empty_or_other(single_record):
        """Detect  Rate Limiting Notice or empty records
           so they can be skipped.
        """
        skip = False
        if not single_record or \
            (isinstance(single_record, dict) and
             single_record.get('limit')):
            skip = True
        return skip

    def fetch_json_data_from_lbsn(
            self, cursor, start_id=None, get_max=None,
            number_of_records_to_fetch=10000,
            schema_name=None, table_name=None, key_col=None
    ) -> Optional[List[List[str]]]:
        """Fetches records from Postgres DB

        Keyword arguments:
        cursor -- db-cursor
        start_id -- Offset for querying
        get_max -- optional limit for retrieving records
        number_of_records_to_fetch -- how many records should get fetched
        """
        # if transferlimit is below number_of_records_to_fetch, e.g.  10000,
        # retrieve only necessary volume of records
        if get_max:
            number_of_records_to_fetch = min(
                number_of_records_to_fetch, get_max)
        query_sql = InputSQL.LBSN.get_sql(
            schema_name=schema_name, table_name=table_name,
            start_id=start_id,
            number_of_records_to_fetch=number_of_records_to_fetch,
            key_col=key_col)
        cursor.execute(query_sql)
        records = cursor.fetchall()
        if cursor.rowcount == 0:
            return None
        # update last returned db_row_number
        if key_col == None:
            self.continue_number = records[-1][0]
            if not self.start_number:
                # first returned db_row_number
                self.start_number = records[0][0]
        else:
            self.continue_number = records[-1].get(key_col)
            if not self.start_number:
                # first returned db_row_number
                self.start_number = records[0].get(key_col)
        return records

    def fetch_record_from_file(self, file_handle):
        """Fetches CSV or JSON data (including stacked json) from file"""
        if self.file_format in ['txt', 'csv']:
            record_reader = self.fetch_csv_data_from_file(
                file_handle)
        else:
            sys.exit(f'Format {self.file_format} not supported.')
        # return record pipeline
        for record in record_reader:
            yield record

    def fetch_json_data_from_file(self, file_handle):
        """Read json entries from file.

        Typical form is [{json1},{json2}], if is_stacked_json is True:
        will process stacked jsons in the form of {json1}{json2}

        If is_line_separated_json is true:
        {json1}
        {json2}
        ...
        """
        # records = []
        # Stacked JSON is a simple file with many concatenated jsons, e.g.
        # {json1}{json2} etc.
        if self.is_stacked_json:
            # note: this requires loading file completely first
            # not streaming optimized yet
            for record in HF.json_read_wrapper(
                    HF.decode_stacked(file_handle.read())):
                yield record
        if self.is_line_separated_json:
            # json's separated by line ending
            for line in file_handle:
                record = HF.json_load_wrapper(line, single=True)
                yield record
        else:
            # normal json nesting, e.g.  {{record1},{record2}}
            records = HF.json_load_wrapper(file_handle)
            if records:
                if isinstance(records, list):
                    for record in records:
                        yield record
                else:
                    record = records
                    yield record
            yield None
            # streaming version:
            # start_pos = 0
            # while True:
            #     try:
            #         record = json.load(file_handle)
            #         yield record
            #         return
            #     except json.JSONDecodeError as e:
            #         file_handle.seek(start_pos)
            #         json_str = file_handle.read(e.pos)
            #         record = json.loads(json_str)
            #         start_pos += e.pos
            #         yield record

    def fetch_csv_data_from_file(self, file_handle):
        """Read csv entries from file (either *.txt or *.csv).

        The actual CSV formatting is not setable in config yet.
        There are many specifics, e.g.
        # QUOTE_NONE is used here because media saved from Flickr
        does not contain any quotes ""
        """
        if self.csv_delim is None:
            self.csv_delim = ','
        kwargs = {
            'delimiter':self.csv_delim,
            'quotechar':'"',
            'quoting':csv.QUOTE_NONE
        }
        if self.use_csv_dictreader:
            record_reader = csv.DictReader(f=file_handle, **kwargs)
            return record_reader
        record_reader = csv.reader(file_handle, **kwargs)
        return record_reader

    @staticmethod
    def initialize_connection(
            dbuser_output, dbserveraddress_output,
            dbname_output, dbpassword_output, dbserverport_output,
            readonly: bool = True, dict_cursor: Optional[bool] = None):
        """Establishes connection to DB (Postgres)"""

        if dbuser_output:
            connection = DBConnection(
                serveraddress=dbserveraddress_output,
                dbname=dbname_output,
                user=dbuser_output,
                password=dbpassword_output,
                port=dbserverport_output,
                readonly=readonly)
            conn, cursor = connection.connect(dict_cursor)
        else:
            conn = None
            cursor = None
        return conn, cursor

    @staticmethod
    def _read_local_files(input_path=None, recursive_load=None,
                          local_file_type=None, skip_until_file=None):
        """Read Local Files according to config parameters and
        returns list of file-paths
        """
        if recursive_load:
            excludefolderlist = ["01_DataSetHistory",
                                 "02_UserData", "03_ClippedData", "04_MapVis"]
            excludestartswithfile = ["log", "settings", "GridCoordinates"]
            loc_filelist = \
                LoadData.scan_rec(input_path,
                                  file_format=local_file_type,
                                  excludefolderlist=excludefolderlist,
                                  excludestartswithfile=excludestartswithfile)
        else:
            loc_filelist_gen = input_path.glob(f'*.{local_file_type}')
            loc_filelist = []
            for file_path in loc_filelist_gen:
                loc_filelist.append(file_path)
        if skip_until_file:
            print("Implement skip until file")
        input_count = (len(loc_filelist))
        if input_count == 0:
            sys.exit("No location files found.")
        else:
            return loc_filelist

    @staticmethod
    def load_geocodes(geo_config):
        """Loads coordinates-string tuples for geocoding
        text locations (Optional)
        """
        locationsgeocode_dict = GeocodeLocations()
        locationsgeocode_dict.load_geocodelist(geo_config)
        geocode_dict = locationsgeocode_dict.geocode_dict
        return geocode_dict

    @staticmethod
    def load_ignore_sources(list_source=None):
        """Loads list of source types to be ignored"""
        ignore_source_list = set()
        with open(list_source, newline='', encoding='utf8') as file_handle:
            for ignore_source in file_handle:
                ignore_source_list.add(ignore_source.strip())
        return ignore_source_list

    @staticmethod
    def scan_rec(root, subdirlimit=2, file_format="csv",
                 excludefolderlist=None, excludestartswithfile=None):
        """Recursively scan subdir for datafiles"""
        rval = []
        if excludefolderlist is None:
            excludefolderlist = []
        if excludestartswithfile is None:
            excludestartswithfile = []

        def do_scan(start_dir, output, depth=0):
            for file_handle in os.listdir(start_dir):
                file_handle_path = os.path.join(start_dir, file_handle)
                if os.path.isdir(file_handle_path):
                    if depth < subdirlimit:
                        efound = False
                        # check for excludefolders:
                        for entry in excludefolderlist:
                            if entry in file_handle_path:
                                efound = True
                                break
                        if efound is False:
                            do_scan(file_handle_path, output, depth+1)
                else:
                    if file_handle_path.endswith(file_format):
                        efound = False
                        for entry in excludestartswithfile:
                            if ntpath.basename(
                                    file_handle_path).startswith(entry):
                                efound = True
                                break
                        if efound is False:
                            output.append(file_handle_path)
        do_scan(root, rval, 0)
        return rval



# -*- coding: utf-8 -*-

"""Setup config for cx_freeze (build)
"""


import os.path

from cx_Freeze import Executable, setup

# Derive Package Paths Dynamically
PYTHON_INSTALL_DIR = os.path.dirname(os.path.dirname(os.__file__))

VERSION_DICT = {}
with open("lbsntransform/version.py") as fp:
    exec(fp.read(), VERSION_DICT)  # pylint: disable=W0122
VERSION = VERSION_DICT['__version__']

EXCLUDES_MOD = ['tkinter',
                'matplotlib',
                'IPython',
                'ipykernel',
                'jedi',
                'jinja2',
                'jupyter_client',
                'multiprocessing',
                'scipy',
                'numpy']
PACKAGES_MOD = [
    'lbsnstructure>=0.5.1',
    'protobuf',
    'psycopg2',
    'ppygis3',
    'shapely',
    'emoji',
    'requests',
    'regex'
]
INCLUDE_FOLDERS_FILES = [('scripts/00_TransferAll_Default.sh')
                         ]
BUILD_EXE_OPTIONS = {'include_files': INCLUDE_FOLDERS_FILES,
                     "packages": PACKAGES_MOD, "excludes": EXCLUDES_MOD}
BASE = None
EXECUTABLES = [
    Executable('lbsntransform/__main__.py', base=BASE,
               targetName="lbsntransform.exe")
]
setup(name="lbsntransform",
      version=VERSION,
      description="Location based social network (LBSN) "
                  "data structure format & transfer tool",
      author='Alexander Dunkel',
      url='https://gitlab.vgiscience.de/lbsn/lbsntransform',
      license='GNU GPLv3 or any higher',
      options={'build_exe': BUILD_EXE_OPTIONS},
      executables=EXECUTABLES,
      )



# -*- coding: utf-8 -*-

"""
Module for sql insert functions for LBSN (raw) db
"""

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.tools.helper_functions import HelperFunctions as HF

from lbsntransform.output.lbsn.shared_structure_proto_lbsndb import ProtoLBSNMapping


class LBSNSql():
    """Maps LBSN Types to raw SQL Structure"""

    DB_MAPPING = ProtoLBSNMapping()

    @classmethod
    def type_sql_mapper(cls):
        """Assigns record types to SQL Insert SQLs"""
        type_sql_mapping = {
            lbsn.Origin().DESCRIPTOR.name: cls.origin_insertsql,
            lbsn.Country().DESCRIPTOR.name: cls.country_insertsql,
            lbsn.City().DESCRIPTOR.name: cls.city_insertsql,
            lbsn.Place().DESCRIPTOR.name: cls.place_insertsql,
            lbsn.User().DESCRIPTOR.name: cls.user_insertsql,
            lbsn.UserGroup().DESCRIPTOR.name: cls.usergroup_insertsql,
            lbsn.Post().DESCRIPTOR.name: cls.post_insertsql,
            lbsn.Event().DESCRIPTOR.name: cls.event_insertsql,
            lbsn.PostReaction().DESCRIPTOR.name: cls.postreaction_insertsql,
        }
        return type_sql_mapping

    @staticmethod
    def origin_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.City"""
        insert_sql = \
            f'''
            INSERT INTO social."origin" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id)
            DO UPDATE SET
                name = EXCLUDED.name;
            '''
        return insert_sql

    @staticmethod
    def postreaction_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.PostReaction"""
        insert_sql = \
            f'''
            INSERT INTO topical."post_reaction" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, reaction_guid)
            DO UPDATE SET
                reaction_latlng = COALESCE(
                    NULLIF(EXCLUDED.reaction_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                    topical."post_reaction".reaction_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                user_guid = COALESCE(EXCLUDED.user_guid,
                    topical."post_reaction".user_guid),
                referencedPost_guid = COALESCE(EXCLUDED.referencedPost_guid,
                    topical."post_reaction".referencedPost_guid),
                referencedPostreaction_guid = COALESCE(
                    EXCLUDED.referencedPostreaction_guid,
                    topical."post_reaction".referencedPostreaction_guid),
                reaction_type = COALESCE(NULLIF(EXCLUDED.reaction_type, 'unknown'),
                    topical."post_reaction".reaction_type, 'unknown'),
                reaction_date = COALESCE(EXCLUDED.reaction_date,
                    topical."post_reaction".reaction_date),
                reaction_content = COALESCE(EXCLUDED.reaction_content,
                    topical."post_reaction".reaction_content),
                reaction_like_count = COALESCE(EXCLUDED.reaction_like_count,
                    topical."post_reaction".reaction_like_count),
                user_mentions = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_mentions,
                    topical."post_reaction".user_mentions), ARRAY[]::text[]);
            '''
        return insert_sql

    @staticmethod
    def post_insertsql(values_str: str, record_type):
        """Insert SQL for post values
        Note COALESCE:
        - coalesce will return the first value that is not Null
        - NULLIF(value1, value2) returns null if value1 and value2 match,
            otherwise returns value1
        - combining these allows to prevent overwriting of existing
            with default values
        - if existing values are also Null, a 3rd value can be added to
            specify the final default value (e.g. the one define in
            pgtable default)
        - "default" values in postgres table are only used on insert,
            never on update (upsert)
        """
        insert_sql = \
            f'''
            INSERT INTO topical."post" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, post_guid)
            DO UPDATE SET
                post_latlng = COALESCE(
                    NULLIF(EXCLUDED.post_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                    topical."post".post_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                place_guid = COALESCE(EXCLUDED.place_guid,
                    topical."post".place_guid),
                city_guid = COALESCE(EXCLUDED.city_guid,
                    topical."post".city_guid),
                country_guid = COALESCE(EXCLUDED.country_guid,
                    topical."post".country_guid),
                post_geoaccuracy = COALESCE(NULLIF(EXCLUDED.post_geoaccuracy,'unknown'),
                    topical."post".post_geoaccuracy, 'unknown'),
                user_guid = COALESCE(EXCLUDED.user_guid,
                    topical."post".user_guid),
                post_create_date = COALESCE(EXCLUDED.post_create_date,
                    topical."post".post_create_date),
                post_publish_date = COALESCE(EXCLUDED.post_publish_date,
                    topical."post".post_publish_date),
                post_body = COALESCE(EXCLUDED.post_body,
                    topical."post".post_body),
                post_language = COALESCE(EXCLUDED.post_language,
                    topical."post".post_language),
                user_mentions = COALESCE(EXCLUDED.user_mentions,
                    topical."post".user_mentions),
                hashtags = COALESCE(extensions.mergeArrays(EXCLUDED.hashtags,
                    topical."post".hashtags), ARRAY[]::text[]),
                emoji = COALESCE(
                    extensions.mergeArrays(EXCLUDED.emoji,
                    topical."post".emoji), ARRAY[]::text[]),
                post_like_count = COALESCE(EXCLUDED.post_like_count,
                    topical."post".post_like_count),
                post_comment_count = COALESCE(EXCLUDED.post_comment_count,
                    topical."post".post_comment_count),
                post_views_count = COALESCE(EXCLUDED.post_views_count,
                    topical."post".post_views_count),
                post_title = COALESCE(EXCLUDED.post_title,
                    topical."post".post_title),
                post_thumbnail_url = COALESCE(EXCLUDED.post_thumbnail_url,
                    topical."post".post_thumbnail_url),
                post_url = COALESCE(EXCLUDED.post_url,
                    topical."post".post_url),
                post_type = COALESCE(NULLIF(EXCLUDED.post_type, 'text'),
                    topical."post".post_type, 'text'),
                post_filter = COALESCE(EXCLUDED.post_filter,
                    topical."post".post_filter),
                post_quote_count = COALESCE(EXCLUDED.post_quote_count,
                    topical."post".post_quote_count),
                post_share_count = COALESCE(EXCLUDED.post_share_count,
                    topical."post".post_share_count),
                input_source = COALESCE(EXCLUDED.input_source,
                    topical."post".input_source),
                post_content_license = COALESCE(
                    EXCLUDED.post_content_license,
                        topical."post".post_content_license);
            '''
        return insert_sql

    @staticmethod
    def user_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.User"""
        insert_sql = \
            f'''
            INSERT INTO social."user" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, user_guid)
            DO UPDATE SET
                user_name = COALESCE(EXCLUDED.user_name,
                    social."user".user_name),
                user_fullname = COALESCE(EXCLUDED.user_fullname,
                    social."user".user_fullname),
                follows = GREATEST(COALESCE(
                    EXCLUDED.follows, social."user".follows),
                    COALESCE(social."user".follows, EXCLUDED.follows)),
                followed = GREATEST(COALESCE(
                    EXCLUDED.followed, social."user".followed),
                    COALESCE(social."user".followed, EXCLUDED.followed)),
                group_count = GREATEST(COALESCE(
                    EXCLUDED.group_count, social."user".group_count),
                    COALESCE(social."user".group_count, EXCLUDED.group_count)),
                biography = COALESCE(EXCLUDED.biography,
                    social."user".biography),
                post_count = GREATEST(COALESCE(
                    EXCLUDED.post_count, "user".post_count),
                    COALESCE(social."user".post_count, EXCLUDED.post_count)),
                is_private = COALESCE(EXCLUDED.is_private,
                    social."user".is_private),
                url = COALESCE(EXCLUDED.url, social."user".url),
                is_available = COALESCE(EXCLUDED.is_available,
                    social."user".is_available),
                user_language = COALESCE(EXCLUDED.user_language,
                    social."user".user_language),
                user_location = COALESCE(EXCLUDED.user_location,
                    social."user".user_location),
                user_location_geom = COALESCE(EXCLUDED.user_location_geom,
                    social."user".user_location_geom),
                liked_count = GREATEST(COALESCE(
                    EXCLUDED.liked_count, social."user".liked_count),
                    COALESCE(social."user".liked_count, EXCLUDED.liked_count)),
                active_since = COALESCE(EXCLUDED.active_since,
                    social."user".active_since),
                profile_image_url = COALESCE(EXCLUDED.profile_image_url,
                    social."user".profile_image_url),
                user_timezone = COALESCE(EXCLUDED.user_timezone,
                    social."user".user_timezone),
                user_utc_offset = COALESCE(EXCLUDED.user_utc_offset,
                    social."user".user_utc_offset),
                user_groups_member = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_groups_member,
                    social."user".user_groups_member), ARRAY[]::text[]),
                user_groups_follows = COALESCE(
                    extensions.mergeArrays(EXCLUDED.user_groups_follows,
                    social."user".user_groups_follows), ARRAY[]::text[]);
            '''
        return insert_sql

    @staticmethod
    def usergroup_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.UserGroup"""
        insert_sql = \
            f'''
            INSERT INTO social."user_groups" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, usergroup_guid)
            DO UPDATE SET
                usergroup_name = COALESCE(EXCLUDED.usergroup_name,
                    social."user_groups".usergroup_name),
                usergroup_description = COALESCE(
                    EXCLUDED.usergroup_description,
                    social."user_groups".usergroup_description),
                member_count = GREATEST(COALESCE(
                    EXCLUDED.member_count, social."user_groups".member_count),
                    COALESCE(social."user_groups".member_count,
                    EXCLUDED.member_count)),
                usergroup_createdate = COALESCE(
                    EXCLUDED.usergroup_createdate,
                    social."user_groups".usergroup_createdate),
                user_owner = COALESCE(EXCLUDED.user_owner,
                    social."user_groups".user_owner);
            '''
        # No coalesce for user: in case user changes or
        # removes information, this should also be removed from the record
        return insert_sql

    @staticmethod
    def place_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.Place"""
        insert_sql = \
            f'''
            INSERT INTO spatial."place" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,place_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, spatial."place".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,spatial."place".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    spatial."place".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(
                    NULLIF(EXCLUDED.geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                    spatial."place".geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    spatial."place".geom_area),
                url = COALESCE(EXCLUDED.url, spatial."place".url),
                city_guid = COALESCE(EXCLUDED.city_guid,
                    spatial."place".city_guid),
                post_count = GREATEST(COALESCE(EXCLUDED.post_count,
                    spatial."place".post_count), COALESCE(
                        spatial."place".post_count, EXCLUDED.post_count)),
                place_description = COALESCE(
                    EXCLUDED.place_description, spatial."place".place_description),
                place_website = COALESCE(
                    EXCLUDED.place_website, spatial."place".place_website),
                place_phone = COALESCE(
                    EXCLUDED.place_phone, spatial."place".place_phone),
                address = COALESCE(
                    EXCLUDED.address, spatial."place".address),
                zip_code = COALESCE(
                    EXCLUDED.zip_code, spatial."place".zip_code),
                attributes = COALESCE(
                    EXCLUDED.attributes, spatial."place".attributes),
                checkin_count = COALESCE(
                    EXCLUDED.checkin_count, spatial."place".checkin_count),
                like_count = COALESCE(
                    EXCLUDED.like_count, spatial."place".like_count),
                parent_places = COALESCE(
                    extensions.mergeArrays(EXCLUDED.parent_places,
                    spatial."place".parent_places), ARRAY[]::text[]);
            '''
        return insert_sql

    @staticmethod
    def city_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.City"""
        insert_sql = \
            f'''
            INSERT INTO spatial."city" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,city_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, spatial."city".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,spatial."city".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    spatial."city".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(
                    NULLIF(EXCLUDED.geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                    spatial."city".geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    spatial."city".geom_area),
                url = COALESCE(EXCLUDED.url, spatial."city".url),
                country_guid = COALESCE(EXCLUDED.country_guid,
                    spatial."city".country_guid),
                sub_type = COALESCE(EXCLUDED.sub_type, spatial."city".sub_type);
            '''
        return insert_sql

    @staticmethod
    def country_insertsql(values_str: str, record_type):
        """SQL and value injection for lbsn.Country"""
        insert_sql = \
            f'''
            INSERT INTO spatial."country" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id,country_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name, spatial."country".name),
                name_alternatives = COALESCE((
                    SELECT array_remove(altNamesNewArray,spatial."country".name)
                    from extensions.mergeArrays(EXCLUDED.name_alternatives,
                    spatial."country".name_alternatives) AS altNamesNewArray),
                    ARRAY[]::text[]),
                geom_center = COALESCE(
                    NULLIF(EXCLUDED.geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                    spatial."country".geom_center,
                    '{HF.NULL_GEOM_HEX}'),
                geom_area = COALESCE(EXCLUDED.geom_area,
                    spatial."country".geom_area),
                url = COALESCE(EXCLUDED.url, spatial."country".url);
            '''
        # Array merge of alternatives:
        # Arrays cannot be null, therefore COALESCE(
        # [if array not null],[otherwise create empty array])
        # We don't want the english name to appear in alternatives,
        # therefore: array_remove(altNamesNewArray,"country".name)
        # Finally, merge New Entries with existing ones (distinct):
        # extensions.mergeArrays([new],[old]) uses custom mergeArrays
        # function (see function definitions)
        return insert_sql

    @staticmethod
    def event_insertsql(values_str: str, record_type):
        """Insert SQL for event values
        """
        insert_sql = \
            f'''
            INSERT INTO temporal."event" (
                {LBSNSql.DB_MAPPING.get_header_for_type(record_type)})
            VALUES {values_str}
            ON CONFLICT (origin_id, event_guid)
            DO UPDATE SET
                name = COALESCE(EXCLUDED.name,
                    temporal."event".name),
                event_latlng = COALESCE(
                    NULLIF(EXCLUDED.event_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                    temporal."event".event_latlng,
                    '{HF.NULL_GEOM_HEX}'),
                event_area = COALESCE(EXCLUDED.event_area,
                    temporal."event".event_area),
                event_website = COALESCE(EXCLUDED.event_website,
                    temporal."event".event_website),
                event_date = COALESCE(EXCLUDED.event_date,
                    temporal."event".event_date),
                event_date_start = COALESCE(EXCLUDED.event_date_start,
                    temporal."event".event_date_start),
                event_date_end = COALESCE(EXCLUDED.event_date_end,
                    temporal."event".event_date_end),
                duration = COALESCE(EXCLUDED.duration,
                    temporal."event".duration),
                place_guid = COALESCE(EXCLUDED.place_guid,
                    temporal."event".place_guid),
                city_guid = COALESCE(EXCLUDED.city_guid,
                    temporal."event".city_guid),
                country_guid = COALESCE(EXCLUDED.country_guid,
                    temporal."event".country_guid),
                user_guid = COALESCE(EXCLUDED.user_guid,
                    temporal."event".user_guid),
                event_description = COALESCE(EXCLUDED.event_description,
                    temporal."event".event_description),
                event_type = COALESCE(EXCLUDED.event_type,
                    temporal."event".event_type),
                event_share_count = COALESCE(EXCLUDED.event_share_count,
                    temporal."event".event_share_count),
                event_like_count = COALESCE(EXCLUDED.event_like_count,
                    temporal."event".event_like_count),
                event_comment_count = COALESCE(EXCLUDED.event_comment_count,
                    temporal."event".event_comment_count),
                event_views_count = COALESCE(EXCLUDED.event_views_count,
                    temporal."event".event_views_count),
                event_engage_count = COALESCE(EXCLUDED.event_engage_count,
                    temporal."event".event_engage_count);
            '''
        return insert_sql


# -*- coding: utf-8 -*-

"""LBSNtransform package import specifications"""

from lbsntransform.version import __version__

from lbsntransform.lbsntransform_ import LBSNTransform
from lbsntransform.config.config import BaseConfig
from lbsntransform.input.load_data import LoadData
from lbsntransform.output.hll.shared_structure_proto_hlldb import ProtoHLLMapping
from lbsntransform.output.lbsn.shared_structure_proto_lbsndb import ProtoLBSNMapping
from lbsntransform.output.shared_structure import (
    GeocodeLocations, LBSNRecordDicts, TimeMonitor)
from lbsntransform.tools.helper_functions import HelperFunctions as HF

# pdoc documentation exclude/include format
__pdoc__ = {}

# pdoc documentation include format
__pdoc__["lbsntransform.__main__"] = True
